import copy
import dataclasses
import datetime
import logging
import os
from collections.abc import Mapping, Sequence
from typing import Literal

import dacite
import numpy as np
import torch
import xarray as xr
from xarray.coding.times import CFDatetimeCoder

import fme
import fme.core.logging_utils as logging_utils
from fme.ace.aggregator.inference import InferenceAggregatorConfig
from fme.ace.data_loading.batch_data import BatchData, PrognosticState
from fme.ace.data_loading.getters import get_forcing_data
from fme.ace.data_loading.inference import (
    ExplicitIndices,
    ForcingDataLoaderConfig,
    InferenceInitialConditionIndices,
    TimestampList,
)
from fme.ace.inference.data_writer import DataWriterConfig, PairedDataWriter
from fme.ace.inference.data_writer.dataset_metadata import DatasetMetadata
from fme.ace.stepper import (
    Stepper,
    StepperOverrideConfig,
    load_stepper,
    load_stepper_config,
)
from fme.ace.stepper.single_module import StepperConfig
from fme.core.cli import prepare_config, prepare_directory
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.dataset_info import IncompatibleDatasetInfo
from fme.core.dicts import to_flat_dict
from fme.core.generics.inference import get_record_to_wandb, run_inference
from fme.core.logging_utils import LoggingConfig
from fme.core.timing import GlobalTimer

from .evaluator import resolve_variable_metadata, validate_time_coarsen_config

StartIndices = InferenceInitialConditionIndices | ExplicitIndices | TimestampList


@dataclasses.dataclass
class InitialConditionConfig:
    """
    Configuration for initial conditions.

    .. note::
        The data specified under path should contain a time dimension of at least
        length 1. If multiple times are present in the dataset specified by ``path``,
        the inference will start an ensemble simulation using each IC along a
        leading sample dimension. Specific times can be selected from the dataset
        by using ``start_indices``.

    Parameters:
        path: The path to the initial conditions dataset.
        engine: The engine used to open the dataset.
        start_indices: optional specification of the subset of
            initial conditions to use.
    """

    path: str
    engine: Literal["netcdf4", "h5netcdf", "zarr"] = "netcdf4"
    start_indices: StartIndices | None = None

    def get_dataset(self) -> xr.Dataset:
        ds = xr.open_dataset(
            self.path,
            engine=self.engine,
            decode_times=CFDatetimeCoder(use_cftime=True),
            decode_timedelta=False,
        )
        return self._subselect_initial_conditions(ds)

    def _subselect_initial_conditions(self, ds: xr.Dataset) -> xr.Dataset:
        if self.start_indices is None:
            ic_indices = slice(None, None)
        elif isinstance(self.start_indices, TimestampList):
            time_index = xr.CFTimeIndex(ds.time.values)
            ic_indices = self.start_indices.as_indices(time_index)
        else:
            ic_indices = self.start_indices.as_indices()
        # time is a required variable but not necessarily a dimension
        sample_dim_name = ds.time.dims[0]
        return ds.isel({sample_dim_name: ic_indices})


def get_initial_condition(
    ds: xr.Dataset, prognostic_names: Sequence[str]
) -> PrognosticState:
    """Given a dataset, extract a mapping of variables to tensors.
    and the time coordinate corresponding to the initial conditions.

    Args:
        ds: Dataset containing initial condition data. Must include prognostic_names
            as variables, and they must each have shape (n_samples, n_lat, n_lon).
            Dataset must also include a 'time' variable with length n_samples.
        prognostic_names: Names of prognostic variables to extract from the dataset.

    Returns:
        The initial condition and the time coordinate.
    """
    initial_condition = {}
    for name in prognostic_names:
        if len(ds[name].shape) != 3:
            raise ValueError(
                f"Initial condition variables {name} must have shape "
                f"(n_samples, n_lat, n_lon). Got shape {ds[name].shape}."
            )
        n_samples = ds[name].shape[0]
        initial_condition[name] = torch.tensor(ds[name].values).unsqueeze(dim=1)
    if "time" not in ds:
        raise ValueError("Initial condition dataset must have a 'time' variable.")
    initial_times = xr.DataArray(
        data=ds.time.values[:, None],
        dims=["sample", "time"],
    )
    if initial_times.shape[0] != n_samples:
        raise ValueError(
            "Length of 'time' variable must match first dimension of variables "
            f"in initial condition dataset. Got {initial_times.shape[0]} "
            f"and {n_samples}."
        )

    batch_data = BatchData.new_on_cpu(
        data=initial_condition,
        time=initial_times,
        horizontal_dims=["lat", "lon"],
    )
    return batch_data.get_start(prognostic_names, n_ic_timesteps=1)


@dataclasses.dataclass
class InferenceConfig:
    """
    Configuration for running inference.

    Parameters:
        experiment_dir: Directory to save results to.
        n_forward_steps: Number of steps to run the model forward for.
        checkpoint_path: Path to stepper checkpoint to load.
        logging: Configuration for logging.
        initial_condition: Configuration for initial condition data.
        forcing_loader: Configuration for forcing data.
        forward_steps_in_memory: Number of forward steps to complete in memory
            at a time.
        data_writer: Configuration for data writers.
        aggregator: Configuration for inference aggregator.
        stepper_override: Configuration for overriding select stepper configuration
            options at inference time (optional).
        allow_incompatible_dataset: If True, allow the dataset used for inference
            to be incompatible with the dataset used for stepper training. This should
            be used with caution, as it may allow the stepper to make scientifically
            invalid predictions, but it can allow running inference with incorrectly
            formatted or missing grid information.
    """

    experiment_dir: str
    n_forward_steps: int
    checkpoint_path: str
    logging: LoggingConfig
    initial_condition: InitialConditionConfig
    forcing_loader: ForcingDataLoaderConfig
    forward_steps_in_memory: int = 10
    data_writer: DataWriterConfig = dataclasses.field(
        default_factory=lambda: DataWriterConfig()
    )
    aggregator: InferenceAggregatorConfig = dataclasses.field(
        default_factory=lambda: InferenceAggregatorConfig()
    )
    stepper_override: StepperOverrideConfig | None = None
    allow_incompatible_dataset: bool = False

    def __post_init__(self):
        if self.data_writer.time_coarsen is not None:
            validate_time_coarsen_config(
                self.data_writer.time_coarsen,
                self.forward_steps_in_memory,
                self.n_forward_steps,
            )

    def configure_logging(self, log_filename: str):
        self.logging.configure_logging(self.experiment_dir, log_filename)

    def configure_wandb(
        self, env_vars: dict | None = None, resumable: bool = False, **kwargs
    ):
        config = to_flat_dict(dataclasses.asdict(self))
        self.logging.configure_wandb(
            config=config, env_vars=env_vars, resumable=resumable, **kwargs
        )

    def load_stepper(self) -> Stepper:
        logging.info(f"Loading trained model checkpoint from {self.checkpoint_path}")
        return load_stepper(self.checkpoint_path, self.stepper_override)

    def load_stepper_config(self) -> StepperConfig:
        logging.info(f"Loading trained model checkpoint from {self.checkpoint_path}")
        return load_stepper_config(self.checkpoint_path, self.stepper_override)

    def get_data_writer(
        self,
        n_initial_conditions: int,
        timestep: datetime.timedelta,
        coords: Mapping[str, np.ndarray],
        variable_metadata: Mapping[str, VariableMetadata],
    ) -> PairedDataWriter:
        return self.data_writer.build_paired(
            experiment_dir=self.experiment_dir,
            # each batch contains all samples, for different times
            n_initial_conditions=n_initial_conditions,
            n_timesteps=self.n_forward_steps,
            timestep=timestep,
            variable_metadata=variable_metadata,
            coords=coords,
            dataset_metadata=DatasetMetadata.from_env(),
        )


def main(
    yaml_config: str,
    segments: int | None = None,
    override_dotlist: Sequence[str] | None = None,
):
    config_data = prepare_config(yaml_config, override=override_dotlist)
    config = dacite.from_dict(
        data_class=InferenceConfig,
        data=config_data,
        config=dacite.Config(strict=True),
    )
    prepare_directory(config.experiment_dir, config_data)
    if segments is None:
        with GlobalTimer():
            return run_inference_from_config(config)
    else:
        config.configure_logging(log_filename="inference_out.log")
        run_segmented_inference(config, segments)


def run_inference_from_config(config: InferenceConfig):
    timer = GlobalTimer.get_instance()
    timer.start_outer("inference")
    timer.start("initialization")

    if not os.path.isdir(config.experiment_dir):
        os.makedirs(config.experiment_dir, exist_ok=True)
    config.configure_logging(log_filename="inference_out.log")
    env_vars = logging_utils.retrieve_env_vars()
    beaker_url = logging_utils.log_beaker_url()
    config.configure_wandb(env_vars=env_vars, notes=beaker_url)

    if fme.using_gpu():
        torch.backends.cudnn.benchmark = True

    logging_utils.log_versions()
    logging.info(f"Current device is {fme.get_device()}")

    stepper_config = config.load_stepper_config()
    data_requirements = stepper_config.get_forcing_window_data_requirements(
        n_forward_steps=config.forward_steps_in_memory
    )
    logging.info("Loading initial condition data")
    initial_condition = get_initial_condition(
        config.initial_condition.get_dataset(), stepper_config.prognostic_names
    )
    stepper = config.load_stepper()
    stepper.set_eval()
    logging.info("Initializing forcing data loader")
    data = get_forcing_data(
        config=config.forcing_loader,
        total_forward_steps=config.n_forward_steps,
        window_requirements=data_requirements,
        initial_condition=initial_condition,
        surface_temperature_name=stepper.surface_temperature_name,
        ocean_fraction_name=stepper.ocean_fraction_name,
    )
    if not config.allow_incompatible_dataset:
        try:
            stepper.training_dataset_info.assert_compatible_with(data.dataset_info)
        except IncompatibleDatasetInfo as err:
            raise IncompatibleDatasetInfo(
                "Inference dataset is not compatible with dataset used for stepper "
                "training. Set allow_incompatible_dataset to True to ignore this "
                f"error. The incompatiblity found was: {str(err)}"
            ) from err

    variable_metadata = resolve_variable_metadata(
        dataset_metadata=data.variable_metadata,
        stepper_metadata=stepper.training_variable_metadata,
        stepper_all_names=stepper_config.all_names,
    )
    dataset_info = data.dataset_info.update_variable_metadata(variable_metadata)
    aggregator = config.aggregator.build(
        dataset_info=dataset_info,
        n_timesteps=config.n_forward_steps + stepper.n_ic_timesteps,
        output_dir=config.experiment_dir,
    )

    writer = config.get_data_writer(
        n_initial_conditions=data.n_initial_conditions,
        timestep=data.timestep,
        coords=data.coords,
        variable_metadata=variable_metadata,
    )

    timer.stop()
    logging.info("Starting inference")
    record_logs = get_record_to_wandb(label="inference")
    run_inference(
        predict=stepper.predict_paired,
        data=data,
        writer=writer,
        aggregator=aggregator,
        record_logs=record_logs,
    )

    timer.start("final_writer_flush")
    logging.info("Starting final flush of data writer")
    writer.flush()
    logging.info("Writing reduced metrics to disk in netcdf format.")
    aggregator.flush_diagnostics()
    timer.stop()

    timer.stop_outer("inference")
    total_steps = config.n_forward_steps * data.n_initial_conditions
    inference_duration = timer.get_duration("inference")
    wandb_logging_duration = timer.get_duration("wandb_logging")
    total_steps_per_second = total_steps / (inference_duration - wandb_logging_duration)
    timer.log_durations()
    logging.info(
        "Total steps per second (ignoring wandb logging): "
        f"{total_steps_per_second:.2f} steps/second"
    )
    summary_logs = {
        "total_steps_per_second": total_steps_per_second,
        **timer.get_durations(),
        **aggregator.get_summary_logs(),
    }
    record_logs([summary_logs])


def run_segmented_inference(config: InferenceConfig, segments: int):
    """Run inference in multiple segments.

    Args:
        config: inference configuration to be used for each individual segment. The
            provided initial condition configuration will only be used for the first
            segment.
        segments: total number of segments desired. Only missing segments will be run.

    Note:
        This is useful when running very long simulations or when saving a large
        amount of output data to disk. The simulation outputs will be split across
        multiple folders, each corresponding to one of the segments and labeled by
        the segment number.
    """
    logging.info(
        f"Starting segmented inference with {segments} segments. "
        f"Saving to {config.experiment_dir}."
    )
    config_copy = copy.deepcopy(config)
    original_wandb_name = os.environ.get("WANDB_NAME")
    for segment in range(segments):
        segment_label = f"segment_{segment:04d}"
        segment_dir = os.path.join(config.experiment_dir, segment_label)
        restart_path = os.path.join(segment_dir, "restart.nc")
        if os.path.exists(restart_path):
            logging.info(f"Skipping segment {segment} because it has already been run.")
        else:
            logging.info(f"Running segment {segment}.")
            config_copy.experiment_dir = segment_dir
            if original_wandb_name is not None:
                os.environ["WANDB_NAME"] = f"{original_wandb_name}-{segment_label}"
            with GlobalTimer():
                run_inference_from_config(config_copy)
        config_copy.initial_condition = InitialConditionConfig(
            path=restart_path, engine="netcdf4"
        )



import datetime

import numpy as np
import pytest
import torch
import xarray as xr

from fme.ace.aggregator.inference import InferenceEvaluatorAggregator
from fme.ace.data_loading.batch_data import BatchData, PairedData
from fme.core.coordinates import LatLonCoordinates
from fme.core.dataset_info import DatasetInfo
from fme.core.device import get_device

TIMESTEP = datetime.timedelta(hours=6)


def get_ds_info(nx: int, ny: int) -> DatasetInfo:
    return DatasetInfo(
        horizontal_coordinates=LatLonCoordinates(
            lon=torch.arange(nx),
            lat=torch.arange(ny),
        ),
        timestep=TIMESTEP,
    )


def get_zero_time(shape, dims):
    return xr.DataArray(np.zeros(shape, dtype="datetime64[ns]"), dims=dims)


def test_logs_labels_exist():
    n_sample = 10
    n_time = 22
    nx = 2
    ny = 2
    ds_info = get_ds_info(nx, ny)
    initial_time = get_zero_time(shape=[n_sample, 0], dims=["sample", "time"])

    agg = InferenceEvaluatorAggregator(
        dataset_info=ds_info,
        n_timesteps=n_time,
        initial_time=initial_time,
        record_step_20=True,
        log_video=True,
        log_zonal_mean_images=True,
        normalize=lambda x: dict(x),
        save_diagnostics=False,
    )
    time = xr.DataArray(np.zeros((n_sample, n_time)), dims=["sample", "time"])

    logs = agg.record_batch(
        data=PairedData(
            prediction={
                "a": torch.randn(n_sample, n_time, nx, ny, device=get_device())
            },
            reference={"a": torch.randn(n_sample, n_time, nx, ny, device=get_device())},
            time=time,
        ),
    )
    assert len(logs) == n_time
    expected_step_keys = [
        "mean/forecast_step",
        "mean/weighted_mean_gen/a",
        "mean/weighted_mean_target/a",
        "mean/weighted_rmse/a",
        "mean/weighted_std_gen/a",
        "mean/weighted_bias/a",
        "mean/weighted_grad_mag_percent_diff/a",
        "mean_norm/forecast_step",
        "mean_norm/weighted_mean_gen/a",
        "mean_norm/weighted_mean_target/a",
        "mean_norm/weighted_rmse/a",
        "mean_norm/weighted_std_gen/a",
        "mean_norm/weighted_bias/a",
    ]
    for log in logs:
        for key in expected_step_keys:
            assert key in log, key
        assert len(log) == len(expected_step_keys), set(log).difference(
            expected_step_keys
        )

    summary_logs = agg.get_summary_logs()
    expected_keys = [
        "mean_step_20/loss",
        "mean_step_20/weighted_rmse/a",
        "mean_step_20/weighted_bias/a",
        "mean_step_20/weighted_grad_mag_percent_diff/a",
        "power_spectrum/a",
        "power_spectrum/negative_norm_bias/a",
        "power_spectrum/positive_norm_bias/a",
        "power_spectrum/mean_abs_norm_bias/a",
        "power_spectrum/smallest_scale_norm_bias/a",
        "time_mean/rmse/a",
        "time_mean/bias/a",
        "time_mean/bias_map/a",
        "time_mean/gen_map/a",
        "time_mean_norm/rmse/a",
        "time_mean_norm/gen_map/a",
        "time_mean_norm/rmse/channel_mean",
        "zonal_mean/error/a",
        "zonal_mean/gen/a",
        "video/a",
    ]
    assert set(summary_logs.keys()) == set(expected_keys)


def test_inference_logs_labels_exist():
    n_sample = 10
    n_time = 22
    nx = 2
    ny = 2
    ds_info = get_ds_info(nx, ny)
    initial_time = (get_zero_time(shape=[n_sample, 0], dims=["sample", "time"]),)
    agg = InferenceEvaluatorAggregator(
        dataset_info=ds_info,
        n_timesteps=n_time,
        initial_time=initial_time,
        record_step_20=True,
        log_video=True,
        normalize=lambda x: dict(x),
        save_diagnostics=False,
    )
    logs = agg.record_batch(
        data=PairedData(
            prediction={
                "a": torch.randn(n_sample, n_time, nx, ny, device=get_device())
            },
            reference={"a": torch.randn(n_sample, n_time, nx, ny, device=get_device())},
            time=xr.DataArray(np.zeros((n_sample, n_time)), dims=["sample", "time"]),
        ),
    )
    assert isinstance(logs, list)
    assert len(logs) == n_time
    assert "mean/weighted_bias/a" in logs[0]
    assert "mean/weighted_mean_gen/a" in logs[0]
    assert "mean/weighted_mean_target/a" in logs[0]
    assert "mean/weighted_grad_mag_percent_diff/a" in logs[0]
    assert "mean/weighted_rmse/a" in logs[0]
    assert "mean_norm/weighted_bias/a" in logs[0]
    assert "mean_norm/weighted_mean_gen/a" in logs[0]
    assert "mean_norm/weighted_mean_target/a" in logs[0]
    assert "mean_norm/weighted_rmse/a" in logs[0]
    # series/table data should be rolled out, not included as a table
    assert "mean/series" not in logs[0]
    assert "mean_norm/series" not in logs[0]
    assert "reduced/series" not in logs[0]
    assert "reduced_norm/series" not in logs[0]


@pytest.mark.parametrize(
    "window_len, n_windows",
    [
        pytest.param(3, 1, id="single_window"),
        pytest.param(3, 2, id="two_windows"),
    ],
)
def test_inference_logs_length(window_len: int, n_windows: int):
    """
    Test that the inference logs are the correct length when using one or more
    windows.
    """
    nx, ny = 4, 4
    ds_info = get_ds_info(nx, ny)
    initial_time = (get_zero_time(shape=[2, 0], dims=["sample", "time"]),)
    agg = InferenceEvaluatorAggregator(
        dataset_info=ds_info,
        n_timesteps=window_len * n_windows,
        initial_time=initial_time,
        normalize=lambda x: dict(x),
        save_diagnostics=False,
    )
    target_data = BatchData.new_on_device(
        data={"a": torch.zeros([2, window_len, ny, nx], device=get_device())},
        time=xr.DataArray(np.zeros((2, window_len)), dims=["sample", "time"]),
    )
    i_start = 0
    for i in range(n_windows):
        sample_data = {"a": torch.zeros([2, window_len, ny, nx], device=get_device())}
        for i in range(window_len):
            sample_data["a"][..., i, :, :] = float(i_start + i)
        paired_data = PairedData.new_on_device(
            prediction=sample_data,
            reference=target_data.data,
            time=xr.DataArray(np.zeros((2, window_len)), dims=["sample", "time"]),
        )
        logs = agg.record_batch(
            data=paired_data,
        )
        assert len(logs) == window_len
        i_start += window_len


def test_flush_diagnostics(tmpdir):
    nx, ny, n_sample, n_time = 2, 2, 10, 21
    ds_info = get_ds_info(nx, ny)
    initial_time = get_zero_time(shape=[n_sample, 0], dims=["sample", "time"])
    agg = InferenceEvaluatorAggregator(
        dataset_info=ds_info,
        n_timesteps=n_time,
        initial_time=initial_time,
        normalize=lambda x: dict(x),
        output_dir=tmpdir,
        record_step_20=True,
        log_zonal_mean_images=True,
        log_video=True,
        log_histograms=True,
    )
    target_data = {"a": torch.randn(n_sample, n_time, nx, ny, device=get_device())}
    gen_data = {"a": torch.randn(n_sample, n_time, nx, ny, device=get_device())}
    time = get_zero_time(shape=[n_sample, n_time], dims=["sample", "time"])
    agg.record_batch(
        data=PairedData(
            prediction=gen_data,
            reference=target_data,
            time=time,
        ),
    )
    agg.flush_diagnostics()
    expected_files = [  # note: time-dependent aggregators not tested here
        "mean",
        "mean_norm",
        "mean_step_20",
        "zonal_mean",
        "time_mean",
        "histogram",
        "video",
    ]
    for file in expected_files:
        assert (tmpdir / f"{file}_diagnostics.nc").exists()


def test_agg_raises_without_output_dir():
    ds_info = get_ds_info(nx=2, ny=2)
    with pytest.raises(
        ValueError, match="Output directory must be set to save diagnostics"
    ):
        InferenceEvaluatorAggregator(
            dataset_info=ds_info,
            n_timesteps=1,
            initial_time=get_zero_time(shape=[1, 0], dims=["sample", "time"]),
            normalize=lambda x: dict(x),
            save_diagnostics=True,
            output_dir=None,
        )



import datetime
import logging
from collections.abc import Callable, MutableMapping

import torch

from fme.core.atmosphere_data import AtmosphereData, HasAtmosphereVerticalIntegral
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.ocean_derived_variables import get_ocean_derived_variable_metadata
from fme.core.typing_ import TensorDict

DerivedVariableFunc = Callable[[AtmosphereData, datetime.timedelta], torch.Tensor]


_DERIVED_VARIABLE_REGISTRY: MutableMapping[
    str, tuple[DerivedVariableFunc, VariableMetadata]
] = {}


def get_derived_variable_metadata() -> dict[str, VariableMetadata]:
    return {
        **get_atmosphere_derived_variable_metadata(),
        **get_ocean_derived_variable_metadata(),
    }


def get_atmosphere_derived_variable_metadata() -> dict[str, VariableMetadata]:
    return {
        label: metadata for label, (_, metadata) in _DERIVED_VARIABLE_REGISTRY.items()
    }


def register(metadata: VariableMetadata):
    def decorator(func: DerivedVariableFunc):
        label = func.__name__
        if label in _DERIVED_VARIABLE_REGISTRY:
            raise ValueError(f"Function {label} has already been added to registry.")
        _DERIVED_VARIABLE_REGISTRY[label] = (func, metadata)
        return func

    return decorator


@register(VariableMetadata("Pa", "Surface pressure due to dry air only"))
def surface_pressure_due_to_dry_air(
    data: AtmosphereData,
    timestep: datetime.timedelta,
) -> torch.Tensor:
    return data.surface_pressure_due_to_dry_air


@register(
    VariableMetadata("Pa/s", "Absolute value of tendency of dry air surface pressure")
)
def surface_pressure_due_to_dry_air_absolute_tendency(
    data: AtmosphereData,
    timestep: datetime.timedelta,
) -> torch.Tensor:
    ps_dry = data.surface_pressure_due_to_dry_air
    abs_ps_dry_tendency = torch.zeros_like(ps_dry)
    abs_ps_dry_tendency[:, 1:] = torch.diff(ps_dry, n=1, dim=1).abs()
    return abs_ps_dry_tendency


@register(VariableMetadata("kg/m**2", "Total water path"))
def total_water_path(
    data: AtmosphereData,
    timestep: datetime.timedelta,
) -> torch.Tensor:
    return data.total_water_path


@register(VariableMetadata("kg/m**2/s", "Total water path budget residual"))
def total_water_path_budget_residual(
    data: AtmosphereData,
    timestep: datetime.timedelta,
):
    total_water_path = data.total_water_path
    twp_total_tendency = (total_water_path[:, 1:] - total_water_path[:, :-1]) / (
        timestep.total_seconds()
    )
    twp_budget_residual = torch.zeros_like(total_water_path)
    # no budget residual on initial step
    twp_budget_residual[:, 1:] = twp_total_tendency - (
        data.evaporation_rate[:, 1:]
        - data.precipitation_rate[:, 1:]
        + data.tendency_of_total_water_path_due_to_advection[:, 1:]
    )
    return twp_budget_residual


@register(VariableMetadata("W/m**2", "Net TOA radiative flux into atmosphere"))
def net_energy_flux_toa_into_atmosphere(
    data: AtmosphereData,
    timestep: datetime.timedelta,
):
    return data.net_top_of_atmosphere_energy_flux


@register(VariableMetadata("W/m**2", "Net surface energy flux into atmosphere"))
def net_energy_flux_sfc_into_atmosphere(
    data: AtmosphereData,
    timestep: datetime.timedelta,
):
    # property is defined as positive into surface, but want to compare to
    # MSE tendency defined as positive into atmosphere
    return -data.net_surface_energy_flux


@register(
    VariableMetadata(
        "W/m**2", "Net energy flux through TOA and surface into atmosphere"
    )
)
def net_energy_flux_into_atmospheric_column(
    data: AtmosphereData, timestep: datetime.timedelta
):
    return data.net_energy_flux_into_atmosphere


@register(VariableMetadata("J/m**2", "Total energy path following ACE2 assumptions"))
def total_energy_ace2_path(
    data: AtmosphereData,
    timestep: datetime.timedelta,
):
    return data.total_energy_ace2_path


@register(
    VariableMetadata(
        "W/m**2", "Tendency of total energy path following ACE2 assumptions"
    )
)
def total_energy_ace2_path_tendency(
    data: AtmosphereData,
    timestep: datetime.timedelta,
):
    mse = total_energy_ace2_path(data, timestep)
    mse_tendency = torch.zeros_like(mse)
    mse_tendency[:, 1:] = torch.diff(mse, n=1, dim=1) / timestep.total_seconds()
    return mse_tendency


@register(
    VariableMetadata(
        "W/m**2",
        "Implied advective tendency of total energy path assuming closed budget",
    )
)
def implied_tendency_of_total_energy_ace2_path_due_to_advection(
    data: AtmosphereData,
    timestep: datetime.timedelta,
):
    """Implied tendency of total energy path due to advection.

    This is computed as a residual from the column total energy budget.
    """
    column_energy_tendency = total_energy_ace2_path_tendency(data, timestep)
    flux_through_vertical_boundaries = data.net_energy_flux_into_atmosphere
    implied_column_heating = column_energy_tendency - flux_through_vertical_boundaries
    return implied_column_heating


@register(VariableMetadata("m/s", "Windspeed at 10m above surface"))
def windspeed_at_10m(data: AtmosphereData, timestep: datetime.timedelta):
    return data.windspeed_at_10m


def _compute_derived_variable(
    data: TensorDict,
    vertical_coordinate: HasAtmosphereVerticalIntegral | None,
    timestep: datetime.timedelta,
    label: str,
    derived_variable_func: DerivedVariableFunc,
    forcing_data: TensorDict | None = None,
) -> TensorDict:
    """Computes a derived variable and adds it to the given data.

    The derived variable name must not already exist in the data.

    If any required input data are not available,
    the derived variable will not be computed.

    Args:
        data: dictionary of data add the derived variable to.
        vertical_coordinate: the vertical coordinate.
        timestep: Timestep of the model.
        label: the name of the derived variable.
        derived_variable_func: derived variable function to compute.
        forcing_data: optional dictionary of forcing data needed for some derived
            variables. If necessary forcing inputs are missing, the derived
            variable will not be computed.

    Returns:
        A new data dictionary with the derived variable added.
    """
    if label in data:
        raise ValueError(
            f"Variable {label} already exists. It is not permitted "
            "to overwrite existing variables with derived variables."
        )
    new_data = data.copy()
    if forcing_data is not None:
        for key, value in forcing_data.items():
            if key not in data:
                data[key] = value

    atmosphere_data = AtmosphereData(data, vertical_coordinate)

    try:
        output = derived_variable_func(atmosphere_data, timestep)
    except KeyError as key_error:
        logging.debug(f"Could not compute {label} because {key_error} is missing")
    else:  # if no exception was raised
        new_data[label] = output
    return new_data


def compute_derived_quantities(
    data: TensorDict,
    vertical_coordinate: HasAtmosphereVerticalIntegral | None,
    timestep: datetime.timedelta,
    forcing_data: TensorDict | None = None,
) -> TensorDict:
    """Computes all derived quantities from the given data."""
    for label in _DERIVED_VARIABLE_REGISTRY:
        func = _DERIVED_VARIABLE_REGISTRY[label][0]
        data = _compute_derived_variable(
            data,
            vertical_coordinate,
            timestep,
            label,
            func,
            forcing_data=forcing_data,
        )
    return data



from .testing import mock_distributed, mock_wandb
from .training_history import TrainingHistory, TrainingJob, get_job_id
from .wandb import WandB


def test_training_history_append_extend():
    training_history = TrainingHistory()

    job1 = TrainingJob(
        git_sha="abc123",
        job_id="run_456",
    )

    # Append the job to the history
    training_history.append(job1)

    # Check if the job is appended correctly
    assert len(training_history) == 1
    assert training_history[0].git_sha == "abc123"
    assert training_history[0].job_id == "run_456"

    # Extend the history with another TrainingHistory instances
    job2 = TrainingJob(
        git_sha="def456",
        job_id="run_789",
    )
    TrainingHistory2 = TrainingHistory([job2])
    training_history.extend(TrainingHistory2)

    # Check if the job is appended correctly
    assert len(training_history) == 2
    assert training_history[1].git_sha == "def456"
    assert training_history[1].job_id == "run_789"

    # Get the history of a specific key
    training_history_by_key = training_history.get_history_by_key()
    assert training_history_by_key["git_sha"] == ["abc123", "def456"]
    assert training_history_by_key["job_id"] == ["run_456", "run_789"]


def test_stepper_training_metadata_roundtrip():
    history = TrainingHistory()
    job1 = TrainingJob(
        git_sha="abc123",
        job_id="run_456",
    )
    history.append(job1)

    # roundtrip the training history
    new_history = TrainingHistory.from_state(history.get_state())
    assert new_history == history


def test_get_job_id_wandb():
    with mock_wandb(), mock_distributed():
        wandb = WandB.get_instance()
        assert get_job_id() is None, "wandb is not enabled/configured"
        wandb.configure(log_to_wandb=True)
        wandb.init()
        job_id = get_job_id()
        assert wandb.get_id() == job_id



import logging
from collections.abc import Callable, Iterator
from typing import Generic, Protocol, TypeVar

import torch

from fme.core.generics.aggregator import InferenceAggregatorABC, InferenceLogs
from fme.core.generics.data import InferenceDataABC
from fme.core.generics.writer import NullDataWriter, WriterABC
from fme.core.timing import GlobalTimer
from fme.core.wandb import WandB

PS = TypeVar("PS")  # prognostic state
FD = TypeVar("FD", contravariant=True)  # forcing data
SD = TypeVar("SD", covariant=True)  # stepped data


class PredictFunction(Protocol, Generic[PS, FD, SD]):
    def __call__(
        self,
        initial_condition: PS,
        forcing: FD,
        compute_derived_variables: bool = False,
    ) -> tuple[SD, PS]: ...


class Looper(Generic[PS, FD, SD]):
    """
    Class for stepping a model forward arbitarily many times.
    """

    def __init__(
        self,
        predict: PredictFunction[PS, FD, SD],
        data: InferenceDataABC[PS, FD],
    ):
        """
        Args:
            predict: The prediction function to use.
            data: The data to use.
        """
        self._predict = predict
        self._prognostic_state = data.initial_condition
        self._len = len(data.loader)
        self._loader = iter(data.loader)

    def __iter__(self) -> Iterator[SD]:
        return self

    def __len__(self) -> int:
        return self._len

    def __next__(self) -> SD:
        """Return predictions for the time period corresponding to the next batch
        of forcing data. Also returns the forcing data.
        """
        timer = GlobalTimer.get_instance()
        with timer.context("data_loading"):
            try:
                forcing_data = next(self._loader)
            except StopIteration:
                raise StopIteration
        output_data, self._prognostic_state = self._predict(
            self._prognostic_state,
            forcing=forcing_data,
            compute_derived_variables=True,
        )
        return output_data

    def get_prognostic_state(self) -> PS:
        return self._prognostic_state


def get_record_to_wandb(label: str = "") -> Callable[[InferenceLogs], None]:
    wandb = WandB.get_instance()
    step = 0

    def record_logs(logs: InferenceLogs):
        nonlocal step
        for j, log in enumerate(logs):
            if len(log) > 0:
                if label != "":
                    log = {f"{label}/{k}": v for k, v in log.items()}
                wandb.log(log, step=step + j)
        step += len(logs)

    return record_logs


def run_inference(
    predict: PredictFunction[PS, FD, SD],
    data: InferenceDataABC[PS, FD],
    aggregator: InferenceAggregatorABC[PS, SD],
    writer: WriterABC[PS, SD] | None = None,
    record_logs: Callable[[InferenceLogs], None] | None = None,
):
    """Run extended inference loop given initial condition and forcing data.

    Args:
        predict: The prediction function to use.
        data: Provides an initial condition and appropriately aligned windows of
            forcing data.
        aggregator: Aggregator for collecting and reducing metrics.
        writer: Data writer for saving the inference results to disk.
        record_logs: Function for recording logs. By default, logs are recorded to
            wandb.
    """
    if record_logs is None:
        record_logs = get_record_to_wandb(label="inference")
    if writer is None:
        writer = NullDataWriter()
    timer = GlobalTimer.get_instance()
    with torch.no_grad():
        looper = Looper(predict=predict, data=data)
        with timer.context("aggregator"):
            logs = aggregator.record_initial_condition(
                initial_condition=data.initial_condition,
            )
        with timer.context("wandb_logging"):
            record_logs(logs)
        with timer.context("data_writer"):
            writer.write(data.initial_condition, "initial_condition.nc")
        n_windows = len(looper)
        for i, batch in enumerate(looper):
            logging.info(
                f"Inference: processing output from window {i + 1} of {n_windows}."
            )
            with timer.context("data_writer"):
                writer.append_batch(
                    batch=batch,
                )
            with timer.context("aggregator"):
                logs = aggregator.record_batch(
                    data=batch,
                )
            with timer.context("wandb_logging"):
                record_logs(logs)
        with timer.context("data_writer"):
            prognostic_state = looper.get_prognostic_state()
            writer.write(prognostic_state, "restart.nc")



import dataclasses
import pathlib
from collections.abc import Iterable, Mapping
from copy import copy

import fsspec
import numpy as np
import torch
import torch.jit
import xarray as xr

from fme.core.device import move_tensordict_to_device
from fme.core.typing_ import TensorDict, TensorMapping


@dataclasses.dataclass
class NormalizationConfig:
    """
    Configuration for normalizing data.

    Either global_means_path and global_stds_path or explicit means and stds
    must be provided.

    Parameters:
        global_means_path: Path to a netCDF file containing global means.
        global_stds_path: Path to a netCDF file containing global stds.
        means: Mapping from variable names to means.
        stds: Mapping from variable names to stds.
        fill_nans_on_normalize: Whether to fill NaNs during normalization. If
            true, on normalization NaNs in the denormalized input become zeros in
            the normalized output.
        fill_nans_on_denormalize: Whether to fill NaNs during denormalization. If
            true, on denormalization NaNs in the normalized input become global means in
            the denormalized output.
    """

    global_means_path: str | pathlib.Path | None = None
    global_stds_path: str | pathlib.Path | None = None
    means: Mapping[str, float] = dataclasses.field(default_factory=dict)
    stds: Mapping[str, float] = dataclasses.field(default_factory=dict)
    fill_nans_on_normalize: bool = False
    fill_nans_on_denormalize: bool = False

    def __post_init__(self):
        using_path = (
            self.global_means_path is not None and self.global_stds_path is not None
        )
        using_explicit = len(self.means) > 0 and len(self.stds) > 0
        if using_path and using_explicit:
            raise ValueError(
                "Cannot use both global_means_path and global_stds_path "
                "and explicit means and stds."
            )
        if not (using_path or using_explicit):
            raise ValueError(
                "Must use either global_means_path and global_stds_path "
                "or explicit means and stds."
            )

    def load(self):
        """
        Load the normalization configuration from the netCDF files.

        Updates the configuration so it no longer requires external files.
        """
        if self.global_means_path is not None and self.global_stds_path is not None:
            # convert to explicit means and stds so if the object is stored
            # and reloaded, we no longer need the netCDF files
            means = load_dict_from_netcdf(
                self.global_means_path,
                names=None,
                defaults={"x": 0.0, "y": 0.0, "z": 0.0},
            )
            stds = load_dict_from_netcdf(
                self.global_stds_path,
                names=None,
                defaults={"x": 1.0, "y": 1.0, "z": 1.0},
            )
            self.means = means
            self.stds = stds
            self.global_means_path = None
            self.global_stds_path = None

    def build(self, names: list[str]):
        using_path = (
            self.global_means_path is not None and self.global_stds_path is not None
        )
        if using_path:
            return get_normalizer(
                global_means_path=self.global_means_path,
                global_stds_path=self.global_stds_path,
                names=names,
                fill_nans_on_normalize=self.fill_nans_on_normalize,
                fill_nans_on_denormalize=self.fill_nans_on_denormalize,
            )
        else:
            means = {k: torch.tensor(self.means[k]) for k in names}
            stds = {k: torch.tensor(self.stds[k]) for k in names}
            return StandardNormalizer(
                means=means,
                stds=stds,
                fill_nans_on_normalize=self.fill_nans_on_normalize,
                fill_nans_on_denormalize=self.fill_nans_on_denormalize,
            )


class StandardNormalizer:
    """
    Responsible for normalizing tensors.
    """

    def __init__(
        self,
        means: TensorDict,
        stds: TensorDict,
        fill_nans_on_normalize: bool = False,
        fill_nans_on_denormalize: bool = False,
    ):
        self.means = move_tensordict_to_device(means)
        self.stds = move_tensordict_to_device(stds)
        self._names = set(means).intersection(stds)
        self._fill_nans_on_normalize = fill_nans_on_normalize
        self._fill_nans_on_denormalize = fill_nans_on_denormalize

    @property
    def fill_nans_on_normalize(self):
        return self._fill_nans_on_normalize

    @property
    def fill_nans_on_denormalize(self):
        return self._fill_nans_on_denormalize

    def normalize(self, tensors: TensorMapping) -> TensorDict:
        filtered_tensors = {k: v for k, v in tensors.items() if k in self._names}
        return _normalize(
            filtered_tensors,
            means=self.means,
            stds=self.stds,
            fill_nans=self._fill_nans_on_normalize,
        )

    def denormalize(self, tensors: TensorMapping) -> TensorDict:
        filtered_tensors = {k: v for k, v in tensors.items() if k in self._names}
        return _denormalize(
            filtered_tensors,
            means=self.means,
            stds=self.stds,
            fill_nans=self._fill_nans_on_denormalize,
        )

    def get_state(self):
        """
        Returns state as a serializable data structure.
        """
        return {
            "means": {k: float(v.cpu().numpy().item()) for k, v in self.means.items()},
            "stds": {k: float(v.cpu().numpy().item()) for k, v in self.stds.items()},
            "fill_nans_on_normalize": self._fill_nans_on_normalize,
            "fill_nans_on_denormalize": self._fill_nans_on_denormalize,
        }

    @classmethod
    def from_state(cls, state) -> "StandardNormalizer":
        """
        Loads state from a serializable data structure.
        """
        means = {
            k: torch.tensor(v, dtype=torch.float) for k, v in state["means"].items()
        }
        stds = {k: torch.tensor(v, dtype=torch.float) for k, v in state["stds"].items()}
        return cls(
            means=means,
            stds=stds,
            fill_nans_on_normalize=state.get("fill_nans_on_normalize", False),
            fill_nans_on_denormalize=state.get("fill_nans_on_denormalize", False),
        )

    def get_normalization_config(self) -> NormalizationConfig:
        return NormalizationConfig(
            means={k: float(v.cpu().numpy().item()) for k, v in self.means.items()},
            stds={k: float(v.cpu().numpy().item()) for k, v in self.stds.items()},
            fill_nans_on_normalize=self.fill_nans_on_normalize,
            fill_nans_on_denormalize=self.fill_nans_on_denormalize,
        )


@torch.jit.script
def _normalize(
    tensors: TensorDict,
    means: TensorDict,
    stds: TensorDict,
    fill_nans: bool,
) -> TensorDict:
    normalized = {k: (t - means[k]) / stds[k] for k, t in tensors.items()}
    if fill_nans:
        for k, v in normalized.items():
            normalized[k] = torch.where(torch.isnan(v), torch.zeros_like(v), v)
    return normalized


@torch.jit.script
def _denormalize(
    tensors: TensorDict,
    means: TensorDict,
    stds: TensorDict,
    fill_nans: bool,
) -> TensorDict:
    denormalized = {k: t * stds[k] + means[k] for k, t in tensors.items()}
    if fill_nans:
        for k, v in denormalized.items():
            denormalized[k] = torch.where(
                torch.isnan(v), torch.full_like(v, fill_value=means[k]), v
            )
    return denormalized


def get_normalizer(
    global_means_path, global_stds_path, names: list[str], **normalizer_kwargs
) -> StandardNormalizer:
    means = load_dict_from_netcdf(
        global_means_path, names, defaults={"x": 0.0, "y": 0.0, "z": 0.0}
    )
    means = {k: torch.as_tensor(v, dtype=torch.float) for k, v in means.items()}
    stds = load_dict_from_netcdf(
        global_stds_path, names, defaults={"x": 1.0, "y": 1.0, "z": 1.0}
    )
    stds = {k: torch.as_tensor(v, dtype=torch.float) for k, v in stds.items()}
    return StandardNormalizer(means=means, stds=stds, **normalizer_kwargs)


def load_dict_from_netcdf(
    path: str | pathlib.Path,
    names: Iterable[str] | None,
    defaults: Mapping[str, float | np.ndarray],
) -> dict[str, float]:
    """
    Load a dictionary of scalar variables from a netCDF file.

    Args:
        path: Path to the netCDF file.
        names: List of variable names to load. If None, all variables in the netCDF
            file are loaded.
        defaults: Dictionary of default values for each variable, if not found
            in the netCDF file.
    """
    with fsspec.open(path, "rb") as f:
        ds = xr.load_dataset(f, mask_and_scale=False)

    result = {}
    if names is None:
        names = set(ds.variables.keys()).union(defaults.keys())
        skip_non_scalar = True
    else:
        skip_non_scalar = False
    for c in names:
        if c in ds.variables:
            if skip_non_scalar and ds.variables[c].ndim > 0:
                continue
            result[c] = float(ds.variables[c].values.item())
        elif c in defaults:
            result[c] = float(defaults[c])
        else:
            raise ValueError(f"Variable {c} not found in {path}")
    ds.close()
    return result


def _combine_normalizers(
    base_normalizer: StandardNormalizer,
    override_normalizer: StandardNormalizer,
) -> StandardNormalizer:
    """
    Combine two normalizers by overwriting the base normalizer values that are
    present in the override normalizer.

    NaN-filling behavior is inherited from the base normalizer.
    """
    means, stds = copy(base_normalizer.means), copy(base_normalizer.stds)
    means.update(override_normalizer.means)
    stds.update(override_normalizer.stds)
    return StandardNormalizer(
        means=means,
        stds=stds,
        fill_nans_on_normalize=base_normalizer.fill_nans_on_normalize,
        fill_nans_on_denormalize=base_normalizer.fill_nans_on_denormalize,
    )


@dataclasses.dataclass
class NetworkAndLossNormalizationConfig:
    """
    Combined configuration for network and loss normalization.

    Allows loss normalization to be defined as equal to the network
    normalization, apart from a set of residual-scaled variables.

    Parameters:
        network: The normalization configuration for the network.
        loss: The normalization configuration for the loss. Default is to
            use the network configuration, except for residual-scaled variables
            which instead use the residual configuration if given.
        residual: The normalization configuration for residuals. Cannot be
            provided if loss normalization is also provided.
    """

    network: NormalizationConfig
    loss: NormalizationConfig | None = None
    residual: NormalizationConfig | None = None

    def __post_init__(self):
        if self.loss is not None and self.residual is not None:
            raise ValueError("Cannot provide both loss and residual normalization.")

    def get_network_normalizer(self, names: list[str]) -> StandardNormalizer:
        return self.network.build(names=names)

    def get_loss_normalizer(
        self,
        names: list[str],
        residual_scaled_names: list[str],
    ) -> StandardNormalizer:
        if self.loss is not None:
            return self.loss.build(names=names)
        elif self.residual is not None:
            return _combine_normalizers(
                base_normalizer=self.network.build(names=names),
                override_normalizer=self.residual.build(names=residual_scaled_names),
            )
        else:
            return self.network.build(names=names)

    def load(self):
        self.network.load()
        if self.loss is not None:
            self.loss.load()
        if self.residual is not None:
            self.residual.load()



import dataclasses
import datetime
import pathlib
import tempfile
import unittest
import unittest.mock
from collections.abc import Callable

import dacite
import pytest
import torch
import torch._dynamo.exc
from torch import nn

import fme
from fme.ace.registry.stochastic_sfno import NoiseConditionedSFNOBuilder
from fme.ace.testing.fv3gfs_data import get_scalar_dataset
from fme.core.coordinates import HybridSigmaPressureCoordinate, LatLonCoordinates
from fme.core.corrector.atmosphere import AtmosphereCorrectorConfig, EnergyBudgetConfig
from fme.core.dataset_info import DatasetInfo
from fme.core.distributed import DummyWrapper
from fme.core.multi_call import MultiCallConfig
from fme.core.normalizer import NetworkAndLossNormalizationConfig, NormalizationConfig
from fme.core.registry import ModuleSelector
from fme.core.step.multi_call import MultiCallStepConfig
from fme.core.step.single_module import SingleModuleStepConfig
from fme.core.step.step import StepABC, StepSelector
from fme.core.typing_ import TensorDict

from .radiation import SeparateRadiationStepConfig

DEFAULT_IMG_SHAPE = (45, 90)


def get_network_and_loss_normalization_config(
    names: list[str],
    dir: pathlib.Path | None = None,
) -> NetworkAndLossNormalizationConfig:
    if dir is None:
        return NetworkAndLossNormalizationConfig(
            network=NormalizationConfig(
                means={name: 0.0 for name in names},
                stds={name: 1.0 for name in names},
            ),
        )
    else:
        return NetworkAndLossNormalizationConfig(
            network=NormalizationConfig(
                global_means_path=dir / "means.nc",
                global_stds_path=dir / "stds.nc",
            ),
        )


def get_separate_radiation_config(
    dir: pathlib.Path | None = None,
) -> SeparateRadiationStepConfig:
    normalization = get_network_and_loss_normalization_config(
        names=[
            "prog_a",
            "prog_b",
            "forcing_shared",
            "forcing_rad",
            "diagnostic_rad",
            "diagnostic_main",
        ],
        dir=dir,
    )

    return SeparateRadiationStepConfig(
        builder=ModuleSelector(
            type="SphericalFourierNeuralOperatorNet",
            config={
                "scale_factor": 1,
                "embed_dim": 4,
                "num_layers": 2,
            },
        ),
        radiation_builder=ModuleSelector(
            type="SphericalFourierNeuralOperatorNet",
            config={
                "scale_factor": 1,
                "embed_dim": 4,
                "num_layers": 2,
            },
        ),
        main_prognostic_names=["prog_a", "prog_b"],
        shared_forcing_names=["forcing_shared"],
        radiation_only_forcing_names=["forcing_rad"],
        radiation_diagnostic_names=["diagnostic_rad"],
        main_diagnostic_names=["diagnostic_main"],
        normalization=normalization,
    )


def get_separate_radiation_selector(
    dir: pathlib.Path | None = None,
) -> StepSelector:
    return StepSelector(
        type="separate_radiation",
        config=dataclasses.asdict(get_separate_radiation_config(dir)),
    )


def get_single_module_selector(
    dir: pathlib.Path | None = None,
) -> StepSelector:
    normalization = get_network_and_loss_normalization_config(
        names=[
            "forcing_shared",
            "forcing_rad",
            "diagnostic_main",
            "diagnostic_rad",
        ],
        dir=dir,
    )
    return StepSelector(
        type="single_module",
        config=dataclasses.asdict(
            SingleModuleStepConfig(
                builder=ModuleSelector(
                    type="SphericalFourierNeuralOperatorNet",
                    config={
                        "scale_factor": 1,
                        "embed_dim": 4,
                        "num_layers": 2,
                    },
                ),
                in_names=["forcing_shared", "forcing_rad"],
                out_names=["diagnostic_main", "diagnostic_rad"],
                normalization=normalization,
            ),
        ),
    )


def get_single_module_noise_conditioned_selector(
    dir: pathlib.Path | None = None,
) -> StepSelector:
    normalization = get_network_and_loss_normalization_config(
        names=[
            "forcing_shared",
            "forcing_rad",
            "diagnostic_main",
            "diagnostic_rad",
        ],
        dir=dir,
    )
    return StepSelector(
        type="single_module",
        config=dataclasses.asdict(
            SingleModuleStepConfig(
                builder=ModuleSelector(
                    type="NoiseConditionedSFNO",
                    config=dataclasses.asdict(
                        NoiseConditionedSFNOBuilder(
                            embed_dim=4,
                            noise_embed_dim=4,
                            noise_type="isotropic",
                            num_layers=2,
                            local_blocks=[0],
                        )
                    ),
                ),
                in_names=["forcing_shared", "forcing_rad"],
                out_names=["diagnostic_main", "diagnostic_rad"],
                normalization=normalization,
            ),
        ),
    )


def get_single_module_with_atmosphere_corrector_selector(
    dir: pathlib.Path | None = None,
) -> StepSelector:
    in_names = [
        "DSWRFtoa",
        "HGTsfc",
        "air_temperature_0",
        "air_temperature_1",
        "air_temperature_2",
        "air_temperature_3",
        "air_temperature_4",
        "air_temperature_5",
        "specific_total_water_0",
        "specific_total_water_1",
        "specific_total_water_2",
        "specific_total_water_3",
        "specific_total_water_4",
        "specific_total_water_5",
        "PRESsfc",
        "PRATEsfc",
    ]
    out_names = [
        "air_temperature_0",
        "air_temperature_1",
        "air_temperature_2",
        "air_temperature_3",
        "air_temperature_4",
        "air_temperature_5",
        "specific_total_water_0",
        "specific_total_water_1",
        "specific_total_water_2",
        "specific_total_water_3",
        "specific_total_water_4",
        "specific_total_water_5",
        "PRESsfc",
        "PRATEsfc",
        "LHTFLsfc",
        "SHTFLsfc",
        "ULWRFsfc",
        "ULWRFtoa",
        "DLWRFsfc",
        "DSWRFsfc",
        "USWRFtoa",
        "USWRFsfc",
        "tendency_of_total_water_path_due_to_advection",
    ]
    normalization = get_network_and_loss_normalization_config(
        names=list(set(in_names).union(out_names)),
        dir=dir,
    )
    return StepSelector(
        type="single_module",
        config=dataclasses.asdict(
            SingleModuleStepConfig(
                builder=ModuleSelector(
                    type="SphericalFourierNeuralOperatorNet",
                    config={
                        "scale_factor": 1,
                        "embed_dim": 4,
                        "num_layers": 2,
                    },
                ),
                in_names=in_names,
                out_names=out_names,
                normalization=normalization,
                corrector=AtmosphereCorrectorConfig(
                    conserve_dry_air=True,
                    zero_global_mean_moisture_advection=True,
                    moisture_budget_correction="advection_and_precipitation",
                    force_positive_names=["PRATEsfc"],
                    total_energy_budget_correction=EnergyBudgetConfig(
                        "constant_temperature"
                    ),
                ),
            ),
        ),
    )


def get_multi_call_selector(
    dir: pathlib.Path | None = None,
) -> StepSelector:
    return StepSelector(
        type="multi_call",
        config=dataclasses.asdict(
            MultiCallStepConfig(
                wrapped_step=StepSelector(
                    type="separate_radiation",
                    config=dataclasses.asdict(get_separate_radiation_config(dir)),
                ),
                config=MultiCallConfig(
                    forcing_name="forcing_rad",
                    forcing_multipliers={"double": 2.0},
                    output_names=["diagnostic_rad"],
                ),
            ),
        ),
    )


SEPARATE_RADIATION_CONFIG = get_separate_radiation_config()

SELECTOR_GETTERS = [
    get_single_module_with_atmosphere_corrector_selector,
    get_separate_radiation_selector,
    get_single_module_selector,
    get_single_module_noise_conditioned_selector,
    get_multi_call_selector,
]

SELECTOR_CONFIG_CASES = [
    pytest.param(
        getter(),
        id=getter.__name__,
    )
    for getter in SELECTOR_GETTERS
]

EXPORTABLE_SELECTOR_CONFIG_CASES = [
    # Some configs use th.DiscreteContinuousConvS2 layers
    # which currently do not support export, see
    # https://github.com/NVIDIA/torch-harmonics/issues/73
    pytest.param(
        getter(),
        id=getter.__name__,
    )
    for getter in [
        get_single_module_with_atmosphere_corrector_selector,
        get_separate_radiation_selector,
        get_single_module_selector,
        get_multi_call_selector,
    ]
]

HAS_NEXT_STEP_FORCING_NAME_CASES = [
    pytest.param(
        StepSelector(
            type="separate_radiation",
            config=dataclasses.asdict(SEPARATE_RADIATION_CONFIG),
        ),
        id="multi_call_separate_radiation",
    ),
]

HAS_NEXT_STEP_FORCING_NAME_CASES = [
    pytest.param(
        StepSelector(
            type="separate_radiation",
            config=dataclasses.asdict(SEPARATE_RADIATION_CONFIG),
        ),
        id="separate_radiation",
    ),
]
TIMESTEP = datetime.timedelta(hours=6)


def get_tensor_dict(
    names: list[str], img_shape: tuple[int, int], n_samples: int
) -> TensorDict:
    data_dict = {}
    device = fme.get_device()
    for name in names:
        data_dict[name] = torch.rand(
            n_samples,
            *img_shape,
            device=device,
        )
    return data_dict


def get_step(
    selector: StepSelector,
    img_shape: tuple[int, int],
    init_weights: Callable[[list[nn.Module]], None] = lambda _: None,
) -> StepABC:
    device = fme.get_device()
    horizontal_coordinate = LatLonCoordinates(
        lat=torch.zeros(img_shape[0], device=device),
        lon=torch.zeros(img_shape[1], device=device),
    )
    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=torch.arange(7, device=device), bk=torch.arange(7, device=device)
    )
    dataset_info = DatasetInfo(
        horizontal_coordinates=horizontal_coordinate,
        vertical_coordinate=vertical_coordinate,
        timestep=TIMESTEP,
    )
    return selector.get_step(dataset_info, init_weights)


@pytest.mark.parametrize("config", HAS_NEXT_STEP_FORCING_NAME_CASES)
def test_next_step_forcing_names_is_forcing(config: StepSelector):
    data = dataclasses.asdict(config)
    img_shape = DEFAULT_IMG_SHAPE
    step = get_step(config, img_shape)
    forcing_names = set(step.input_names).difference(step.output_names)
    data["config"]["next_step_forcing_names"] = [list(forcing_names)[0]]
    dacite.from_dict(config.__class__, data, config=dacite.Config(strict=True))


@pytest.mark.parametrize("config", HAS_NEXT_STEP_FORCING_NAME_CASES)
def test_next_step_forcing_names_is_prognostic(config: StepSelector):
    data = dataclasses.asdict(config)
    img_shape = DEFAULT_IMG_SHAPE
    step = get_step(config, img_shape)
    prognostic_names = set(step.output_names).intersection(step.input_names)
    name = list(prognostic_names)[0]
    data["config"]["next_step_forcing_names"] = [name]
    with pytest.raises(ValueError) as err:
        dacite.from_dict(config.__class__, data, config=dacite.Config(strict=True))
    assert "next_step_forcing_name" in str(err.value)
    assert name in str(err.value)


@pytest.mark.parametrize("config", HAS_NEXT_STEP_FORCING_NAME_CASES)
def test_next_step_forcing_names_is_diagnostic(config: StepSelector):
    data = dataclasses.asdict(config)
    img_shape = DEFAULT_IMG_SHAPE
    step = get_step(config, img_shape)
    diagnostic_names = set(step.output_names).difference(step.input_names)
    name = list(diagnostic_names)[0]
    data["config"]["next_step_forcing_names"] = [name]
    with pytest.raises(ValueError) as err:
        dacite.from_dict(config.__class__, data, config=dacite.Config(strict=True))
    assert "next_step_forcing_name" in str(err.value)
    assert name in str(err.value)


@pytest.mark.parametrize("config", SELECTOR_CONFIG_CASES)
def test_step_applies_wrapper(config: StepSelector):
    torch.manual_seed(0)
    img_shape = DEFAULT_IMG_SHAPE
    n_samples = 5
    step = get_step(config, img_shape)
    input_data = get_tensor_dict(step.input_names, img_shape, n_samples)
    next_step_input_data = get_tensor_dict(
        step.next_step_input_names, img_shape, n_samples
    )
    multi_calls = 1
    if isinstance(config._step_config_instance, MultiCallStepConfig):
        if config._step_config_instance.config is not None:
            multi_calls += len(config._step_config_instance.config.forcing_multipliers)

    wrapper = unittest.mock.MagicMock(side_effect=lambda x: x)
    step.step(input_data, next_step_input_data, wrapper=wrapper)
    assert wrapper.call_count == multi_calls * len(step.modules)
    for module in step.modules:
        wrapper.assert_any_call(module)


@pytest.mark.parametrize("config", SELECTOR_CONFIG_CASES)
def test_step_initializes_weights(config: StepSelector):
    torch.manual_seed(0)
    img_shape = DEFAULT_IMG_SHAPE
    init_weights = unittest.mock.MagicMock(side_effect=lambda x: x)
    step = get_step(config, img_shape, init_weights)
    assert init_weights.called
    call_args, call_kwargs = init_weights.call_args
    assert len(call_args) == 1
    assert len(call_kwargs) == 0
    assert isinstance(call_args[0], list | nn.ModuleList)
    assert len(call_args[0]) == len(step.modules)
    for i, module in enumerate(step.modules):
        assert isinstance(module, DummyWrapper)
        assert call_args[0][i] is module.module


@pytest.mark.parametrize("config", EXPORTABLE_SELECTOR_CONFIG_CASES)
def test_export_step(config: StepSelector, very_fast_only: bool):
    if very_fast_only:
        pytest.skip("Skipping non-fast tests")
    torch.manual_seed(0)
    img_shape = DEFAULT_IMG_SHAPE
    n_samples = 5
    step = get_step(config, img_shape)
    input_data = get_tensor_dict(step.input_names, img_shape, n_samples)
    next_step_input_data = get_tensor_dict(
        step.next_step_input_names, img_shape, n_samples
    )
    unscripted_output = step.step(input_data, next_step_input_data)
    for name in step.output_names:
        # informative check for nan values
        torch.testing.assert_close(unscripted_output[name], unscripted_output[name])
    exported = step.export(input_data, next_step_input_data)
    output = exported.module()(input_data, next_step_input_data)
    for name in step.output_names:
        torch.testing.assert_close(output[name], unscripted_output[name])


@pytest.mark.parametrize(
    "get_config",
    SELECTOR_GETTERS,
)
def test_load_config(
    get_config: Callable[[pathlib.Path | None], StepSelector],
):
    non_path_config: StepSelector = get_config(
        None
    )  # doesn't depend on files, use to get names
    all_names = set(non_path_config.input_names).union(non_path_config.output_names)
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = pathlib.Path(temp_dir)
        get_scalar_dataset(all_names, fill_value=0.1).to_netcdf(temp_path / "means.nc")
        get_scalar_dataset(all_names, fill_value=1.1).to_netcdf(temp_path / "stds.nc")
        config = get_config(temp_path)
        config.load()
    img_shape = DEFAULT_IMG_SHAPE
    step = get_step(config, img_shape)
    normalizer = step.normalizer
    assert normalizer.means.keys() == all_names
    assert normalizer.stds.keys() == all_names
    assert all(normalizer.means[name] == 0.1 for name in all_names)
    assert all(normalizer.stds[name] == 1.1 for name in all_names)


@pytest.mark.parametrize(
    "get_config",
    SELECTOR_GETTERS,
)
def test_load_is_required_for_path_config(
    get_config: Callable[[pathlib.Path | None], StepSelector],
):
    non_path_config: StepSelector = get_config(
        None
    )  # doesn't depend on files, use to get names
    all_names = set(non_path_config.input_names).union(non_path_config.output_names)
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = pathlib.Path(temp_dir)
        get_scalar_dataset(all_names, fill_value=0.1).to_netcdf(temp_path / "means.nc")
        get_scalar_dataset(all_names, fill_value=1.1).to_netcdf(temp_path / "stds.nc")
        config = get_config(temp_path)
    img_shape = DEFAULT_IMG_SHAPE
    with pytest.raises(FileNotFoundError):
        get_step(config, img_shape)



import numpy as np
import torch
import xarray as xr

from fme.ace.stepper import TrainOutput
from fme.core.tensors import EnsembleTensorDict
from fme.core.typing_ import TensorDict, TensorMapping


def test_train_output_compute_derived_quantities():
    gen_data = EnsembleTensorDict(
        {
            "a": torch.rand(2, 3, 4, 8),
            "b": torch.rand(2, 3, 4, 8),
        }
    )
    target_data = EnsembleTensorDict(
        {
            "a": torch.rand(2, 1, 4, 8),
            "b": torch.rand(2, 1, 4, 8),
        }
    )

    def derive_func(data: TensorMapping, forcing_data: TensorMapping) -> TensorDict:
        try:
            torch.testing.assert_close(data["a"], gen_data["a"].reshape(2 * 3, 4, 8))
            torch.testing.assert_close(data["b"], gen_data["b"].reshape(2 * 3, 4, 8))
            assert forcing_data["a"].shape == (6, 4, 8)
            # ensemble is fastest dimension between batch and ensemble
            assert (forcing_data["a"][:3] == target_data["a"][0]).all()
            assert (forcing_data["a"][3:] == target_data["a"][1]).all()
        except AssertionError:
            torch.testing.assert_close(data["a"], target_data["a"].reshape(2, 4, 8))
            torch.testing.assert_close(data["b"], target_data["b"].reshape(2, 4, 8))
            assert forcing_data["a"].shape == (2, 4, 8)
            assert (forcing_data["a"] == target_data["a"].reshape(2, 4, 8)).all()
            assert (forcing_data["b"] == target_data["b"].reshape(2, 4, 8)).all()
        return dict(data)

    data = TrainOutput(
        metrics={"loss": torch.tensor(0.0)},
        gen_data=gen_data,
        target_data=target_data,
        time=xr.DataArray(np.zeros((2, 3)), dims=["sample", "time"]),
        normalize=lambda x: x,
        derive_func=derive_func,
    )
    out_data = data.compute_derived_variables()
    for name in gen_data:
        assert name in out_data.gen_data
        assert name in out_data.target_data
        assert out_data.gen_data[name].shape == (2, 3, 4, 8)
        assert out_data.target_data[name].shape == (2, 1, 4, 8)
        assert torch.allclose(out_data.gen_data[name], gen_data[name])
        assert torch.allclose(out_data.target_data[name], target_data[name])



import numpy as np
import pytest
import torch

from fme.ace.aggregator.inference.video import VideoAggregator
from fme.core.device import get_device
from fme.core.typing_ import TensorDict

USING_MPS = get_device().type == "mps"


def _make_data(
    names: list[str], n_samples: int, nx: int, ny: int, offsets: np.ndarray
) -> tuple[TensorDict, TensorDict]:
    """
    Make a 2D array with a pattern of offsets.

    Each variable will have a random horizontal pattern consistent between
    the generated and target data, but with a constant time-dependent offset
    between the two.

    Args:
        names: Names of the variables.
        n_samples: Number of samples in the batch.
        nx: Number of x points.
        ny: Number of y points.
        offsets: 1-D (time-axis) array of offsets, which also determines the
            length of the time axis.
    """
    nt = len(offsets)
    data = np.random.normal(size=(n_samples, nt, len(names), ny, nx))
    gen = {}
    target = {}
    for i_name, name in enumerate(names):
        target[name] = torch.tensor(data[:, :, i_name, :, :], device=get_device())
        gen[name] = torch.tensor(data[:, :, i_name, :, :], device=get_device())
        # add time-variable offset to generated data
        for i_offset, offset in enumerate(offsets):
            gen[name][:, i_offset, :, :] += offset
    return target, gen


def time_select(
    i_start: int,
    i_end: int,
    gen: TensorDict,
    target: TensorDict,
) -> tuple[TensorDict, TensorDict]:
    """
    Select a time range from the generated and target data.

    Args:
        i_start: Start index.
        i_end: End index.
        gen: Generated data.
        target: Target data.
    """
    gen_out = {}
    target_out = {}
    for name in gen.keys():
        gen_out[name] = gen[name][:, i_start:i_end]
        target_out[name] = target[name][:, i_start:i_end]
    return target_out, gen_out


@pytest.mark.parametrize(
    "offsets",
    [
        np.array([0, 1]),
        np.array([0, 1, 2, 3, 4, 5.5]),
    ],
)
@pytest.mark.skipif(USING_MPS, reason="Requires double precision and MPS is enabled.")
def test_video_data(offsets: np.ndarray):
    aggregator = VideoAggregator(n_timesteps=len(offsets), enable_extended_videos=True)
    names = ["a"]
    n_samples = 5
    nx = 8
    ny = 10
    target, gen = _make_data(names, n_samples, nx, ny, offsets)
    n_window_in_memory = 2
    for i_start in range(0, len(offsets) - n_window_in_memory + 1):
        i_end = i_start + n_window_in_memory
        target_window, gen_window = time_select(i_start, i_end, gen, target)
        aggregator.record_batch(
            target_data=target_window, gen_data=gen_window, i_time_start=i_start
        )
    data = aggregator._get_data()
    assert data["bias/a"].target is None
    assert data["min_err/a"].target is None
    assert data["max_err/a"].target is None
    assert data["a"].gen.shape[0] == len(offsets)
    assert data["a"].target.shape[0] == len(offsets)
    assert data["bias/a"].gen.shape[0] == len(offsets)
    assert data["min_err/a"].gen.shape[0] == len(offsets)
    assert data["max_err/a"].gen.shape[0] == len(offsets)
    for i, offset_i in enumerate(offsets):
        np.testing.assert_allclose(
            data["a"].gen[i, :], data["a"].target[i, :] + offset_i
        )
        np.testing.assert_allclose(data["bias/a"].gen[i, :], offset_i)
        np.testing.assert_allclose(data["min_err/a"].gen[i, :], offset_i)
        np.testing.assert_allclose(data["max_err/a"].gen[i, :], offset_i)
        np.testing.assert_allclose(data["gen_var/a"].gen[i, :], 1.0)


@pytest.mark.parametrize(
    "offsets",
    [
        np.array([0, 1]),
        np.array([0, 1, 2, 3, 4, 5.5]),
    ],
)
@pytest.mark.skipif(USING_MPS, reason="Requires double precision and MPS is enabled.")
def test_video_data_without_extended_videos(offsets: np.ndarray):
    aggregator = VideoAggregator(n_timesteps=len(offsets), enable_extended_videos=False)
    names = ["a"]
    n_samples = 5
    nx = 8
    ny = 10
    target, gen = _make_data(names, n_samples, nx, ny, offsets)
    n_window_in_memory = 2
    for i_start in range(0, len(offsets) - n_window_in_memory + 1):
        i_end = i_start + n_window_in_memory
        target_window, gen_window = time_select(i_start, i_end, gen, target)
        aggregator.record_batch(
            target_data=target_window, gen_data=gen_window, i_time_start=i_start
        )
    data = aggregator._get_data()
    assert len(data) == 1
    assert data["a"].gen.shape[0] == len(offsets)
    assert data["a"].target.shape[0] == len(offsets)
    for i, offset_i in enumerate(offsets):
        np.testing.assert_allclose(
            data["a"].gen[i, :], data["a"].target[i, :] + offset_i
        )


def slice_samples(data: TensorDict, i_start: int, i_end: int):
    data_out = {}
    for name in data.keys():
        data_out[name] = data[name][i_start:i_end]
    return data_out


@pytest.mark.parametrize("n_batches", [1, 5])
@pytest.mark.skipif(USING_MPS, reason="Requires double precision and MPS is enabled.")
def test_video_data_values_on_random_inputs(n_batches: int):
    torch.manual_seed(0)
    names = ["a"]
    n_samples = 10
    samples_per_batch = n_samples // n_batches
    nx = 8
    ny = 10
    nt = 7
    aggregator = VideoAggregator(n_timesteps=nt, enable_extended_videos=True)
    offsets = np.zeros([nt])
    target, _ = _make_data(names, n_samples, nx, ny, offsets)
    gen, _ = _make_data(names, n_samples, nx, ny, offsets)
    n_window_in_memory = 2
    for i_start in range(0, len(offsets) - n_window_in_memory + 1):
        i_end = i_start + n_window_in_memory
        target_window, gen_window = time_select(i_start, i_end, gen, target)
        for nb in range(n_batches):  # shouldn't affect results to duplicate batches
            aggregator.record_batch(
                target_data=slice_samples(
                    target_window,
                    i_start=nb * samples_per_batch,
                    i_end=(nb + 1) * samples_per_batch,
                ),
                gen_data=slice_samples(
                    gen_window,
                    i_start=nb * samples_per_batch,
                    i_end=(nb + 1) * samples_per_batch,
                ),
                i_time_start=i_start,
            )
    data = aggregator._get_data()
    assert data["bias/a"].target is None
    assert data["rmse/a"].target is None
    assert data["min_err/a"].target is None
    assert data["max_err/a"].target is None
    assert data["a"].gen.shape[0] == len(offsets)
    assert data["a"].target.shape[0] == len(offsets)
    assert data["bias/a"].gen.shape[0] == len(offsets)
    assert data["rmse/a"].gen.shape[0] == len(offsets)
    assert data["min_err/a"].gen.shape[0] == len(offsets)
    assert data["max_err/a"].gen.shape[0] == len(offsets)
    np.testing.assert_allclose(
        data["a"].gen.cpu().numpy(), gen["a"].mean(dim=0).cpu().numpy()
    )
    np.testing.assert_allclose(
        data["a"].target.cpu().numpy(), target["a"].mean(dim=0).cpu().numpy()
    )
    np.testing.assert_allclose(
        data["bias/a"].gen.cpu().numpy(),
        (gen["a"] - target["a"]).mean(dim=0).cpu().numpy(),
    )
    np.testing.assert_allclose(
        data["rmse/a"].gen.cpu().numpy(),
        ((gen["a"] - target["a"]) ** 2).mean(dim=0).sqrt().cpu().numpy(),
    )
    np.testing.assert_allclose(
        data["min_err/a"].gen.cpu().numpy(),
        (gen["a"] - target["a"]).min(dim=0)[0].cpu().numpy(),
    )
    np.testing.assert_allclose(
        data["max_err/a"].gen.cpu().numpy(),
        (gen["a"] - target["a"]).max(dim=0)[0].cpu().numpy(),
    )
    np.testing.assert_allclose(
        data["gen_var/a"].gen.cpu().numpy(),
        (gen["a"].var(dim=0) / target["a"].var(dim=0)).cpu().numpy(),
    )



from collections.abc import Sequence

import xarray as xr

from fme.core.dataset.concat import XarrayConcat
from fme.core.dataset.properties import DatasetProperties
from fme.core.dataset.subset import XarraySubset
from fme.core.dataset.xarray import XarrayDataset
from fme.core.typing_ import TensorDict


class MergedXarrayDataset:
    def __init__(self, datasets: Sequence[XarrayDataset | XarraySubset | XarrayConcat]):
        self.datasets = datasets

        combined_names = [
            item for dataset in self.datasets for item in dataset[0][0].keys()
        ]
        if len(combined_names) != len(set(combined_names)):
            duplicates = list(
                {item for item in combined_names if combined_names.count(item) > 1}
            )
            raise ValueError(
                f"Variable names must be unique across merged datasets. \
                    \nDuplicates found: {duplicates}"
            )
        for dataset in self.datasets:
            if not dataset.sample_start_times.equals(
                self.datasets[0].sample_start_times
            ):
                raise ValueError(
                    "All datasets in a merged dataset must have the same sample "
                    "start times."
                )
            if not dataset.sample_n_times == self.datasets[0].sample_n_times:
                raise ValueError(
                    "All datasets in the merged datasets \
                         must have the same number of steps per sample item."
                )

    def __getitem__(self, idx: int) -> tuple[TensorDict, xr.DataArray]:
        tensors = {}
        for dataset in self.datasets:
            ds_tensors, time = dataset[idx]
            tensors.update(ds_tensors)
        return tensors, time

    def __len__(self) -> int:
        return len(self.datasets[0])

    def get_sample_by_time_slice(
        self, time_slice: slice
    ) -> tuple[TensorDict, xr.DataArray]:
        tensors: TensorDict = {}
        for dataset in self.datasets:
            ds_tensors, time = dataset.get_sample_by_time_slice(time_slice)
            tensors.update(ds_tensors)
        return tensors, time

    @property
    def all_times(self) -> xr.CFTimeIndex:
        return self.datasets[0].all_times

    @property
    def properties(self) -> DatasetProperties:
        data_properties = None
        for dataset in self.datasets:
            if data_properties is None:
                data_properties = dataset.properties
            else:
                data_properties.update_merged_dataset(dataset.properties)
        if data_properties is None:
            raise ValueError("No dataset available to determine properties")
        return data_properties

    @property
    def total_timesteps(self) -> int:
        return self.datasets[0].total_timesteps



import os

import pytest
import torch

from fme.core.device import get_device
from fme.core.testing import validate_tensor

DIR = os.path.abspath(os.path.dirname(__file__))

from fme.ace.models.ocean.m2lines.samudra import Samudra


@pytest.mark.parametrize("norm", ["batch", "layer", "instance", None, "group"])
def test_samudra_normalization(norm):
    # Model parameters
    input_channels = 4
    output_channels = 3
    batch_size = 2
    height = 64
    width = 64

    # Initialize model
    if norm == "group":
        with pytest.raises(NotImplementedError):
            model = Samudra(
                input_channels=input_channels,
                output_channels=output_channels,
                ch_width=[32, 64],
                dilation=[1, 2],
                n_layers=[1, 1],
                norm=norm,
            )
        return

    model = Samudra(
        input_channels=input_channels,
        output_channels=output_channels,
        ch_width=[32, 64],
        dilation=[1, 2],
        n_layers=[1, 1],
        norm=norm,
    )

    # Create dummy input
    x = torch.randn(batch_size, input_channels, height, width)

    # Forward pass
    output = model(x)

    # Check output shape
    expected_shape = (batch_size, output_channels, height, width)
    assert (
        output.shape == expected_shape
    ), f"Expected output shape {expected_shape}, but got {output.shape}"

    # Check output values
    assert not torch.isnan(output).any(), "Output contains NaN values"
    assert not torch.isinf(output).any(), "Output contains infinite values"


def test_samudra_norm_kwargs():
    model = Samudra(
        input_channels=4,
        output_channels=3,
        ch_width=[32, 64],
        dilation=[1, 2],
        n_layers=[1, 1],
        norm="batch",
        norm_kwargs={"track_running_stats": False},
    )
    for module in model.modules():
        if isinstance(module, torch.nn.BatchNorm2d):
            assert not module.track_running_stats


def test_samudra_output_is_unchanged():
    torch.manual_seed(0)
    input_channels = 2
    output_channels = 3
    img_shape = (9, 18)
    n_samples = 4
    device = get_device()
    model = Samudra(
        input_channels=input_channels,
        output_channels=output_channels,
        ch_width=[3, 3],
        dilation=[1, 2],
        n_layers=[1, 1],
        norm="batch",
    ).to(device)
    # must initialize on CPU to get the same results on GPU
    x = torch.randn(n_samples, input_channels, *img_shape).to(device)
    with torch.no_grad():
        output = model(x)
    assert output.shape == (n_samples, output_channels, *img_shape)
    validate_tensor(
        output,
        os.path.join(DIR, "testdata/test_samudra_output_is_unchanged.pt"),
    )



from collections.abc import Mapping
from typing import Any

import torch
import torch.nn as nn

from .activations import CappedGELU


class BilinearUpsample(torch.nn.Module):
    def __init__(self, upsampling: int = 2, **kwargs):
        super().__init__()
        self.upsampler = torch.nn.Upsample(scale_factor=upsampling, mode="bilinear")

    def forward(self, x):
        return self.upsampler(x)


class AvgPool(torch.nn.Module):
    def __init__(
        self,
        pooling: int = 2,
    ):
        super().__init__()
        self.avgpool = torch.nn.AvgPool2d(pooling)

    def forward(self, x):
        return self.avgpool(x)


class ConvNeXtBlock(torch.nn.Module):
    """
    A convolution block as reported in https://github.com/CognitiveModeling/dlwp-hpx/blob/main/src/dlwp-hpx/dlwp/model/modules/blocks.py.
    This is a modified version of the actual ConvNextblock which
    is used in the HealPix paper.
    """

    def __init__(
        self,
        in_channels: int = 300,
        out_channels: int = 1,
        kernel_size: int = 3,
        dilation: int = 1,
        n_layers: int = 1,
        activation: torch.nn.Module = CappedGELU,
        pad: str = "circular",
        norm: str | None = "instance",
        norm_kwargs: Mapping[str, Any] | None = None,
        upscale_factor: int = 4,
    ):
        super().__init__()
        assert kernel_size % 2 != 0, "Cannot use even kernel sizes!"

        self.N_in = in_channels
        self.N_pad = int((kernel_size + (kernel_size - 1) * (dilation - 1) - 1) / 2)
        self.pad = pad
        self.norm = norm
        self.norm_kwargs = norm_kwargs
        if self.norm_kwargs is None:
            self.norm_kwargs = {}

        assert n_layers == 1, "Can only use a single layer here!"  # Needs fixing

        # 1x1 conv to increase/decrease channel depth if necessary
        if in_channels == out_channels:
            self.skip_module = lambda x: x  # Identity-function required in forward pass
        else:
            self.skip_module = torch.nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=1,
                padding="same",
            )

        # Convolution block
        convblock = []
        convblock.append(
            torch.nn.Conv2d(
                in_channels=in_channels,
                out_channels=int(in_channels * upscale_factor),
                kernel_size=kernel_size,
                dilation=dilation,
            )
        )
        # Batch Norm
        if norm == "batch":
            convblock.append(
                torch.nn.BatchNorm2d(in_channels * upscale_factor, **self.norm_kwargs)
            )
        # Instance Norm
        elif norm == "instance":
            convblock.append(
                torch.nn.InstanceNorm2d(
                    in_channels * upscale_factor, **self.norm_kwargs
                )
            )
        # Layer Norm
        elif norm == "layer":
            convblock.append(
                torch.nn.LayerNorm(in_channels * upscale_factor, **self.norm_kwargs)
            )
        # No Norm
        elif norm is None:
            pass
        else:
            raise NotImplementedError(f"Normalization {norm} not implemented")

        convblock.append(activation())

        convblock.append(
            torch.nn.Conv2d(
                in_channels=int(in_channels * upscale_factor),
                out_channels=int(in_channels * upscale_factor),
                kernel_size=kernel_size,
                dilation=dilation,
            )
        )
        # Batch Norm
        if norm == "batch":
            convblock.append(
                torch.nn.BatchNorm2d(in_channels * upscale_factor, **self.norm_kwargs)
            )
        # Instance Norm
        elif norm == "instance":
            convblock.append(
                torch.nn.InstanceNorm2d(
                    in_channels * upscale_factor, **self.norm_kwargs
                )
            )
        # Layer Norm
        elif norm == "layer":
            convblock.append(
                torch.nn.LayerNorm(in_channels * upscale_factor, **self.norm_kwargs)
            )
        # No Norm
        elif norm is None:
            pass
        else:
            raise NotImplementedError(f"Normalization {norm} not implemented")

        convblock.append(activation())

        # Linear postprocessing
        convblock.append(
            torch.nn.Conv2d(
                in_channels=int(in_channels * upscale_factor),
                out_channels=out_channels,
                kernel_size=1,
                padding="same",
            )
        )
        self.convblock = torch.nn.Sequential(*convblock)

    def forward(self, x):
        skip = self.skip_module(x)
        for layer in self.convblock:
            if isinstance(layer, nn.Conv2d) and layer.kernel_size[0] != 1:
                x = torch.nn.functional.pad(
                    x, (self.N_pad, self.N_pad, 0, 0), mode=self.pad
                )
                x = torch.nn.functional.pad(
                    x, (0, 0, self.N_pad, self.N_pad), mode="constant"
                )
            if isinstance(layer, torch.nn.LayerNorm):
                x = x.permute(0, 2, 3, 1).contiguous()
                x = layer(x)
                x = x.permute(0, 3, 1, 2).contiguous()
            else:
                x = layer(x)
        return skip + x



import dataclasses
import datetime
import logging
import os
from collections.abc import Callable, Mapping, Sequence

import dacite
import numpy as np
import torch

import fme
import fme.core.logging_utils as logging_utils
from fme.ace.aggregator.inference import InferenceEvaluatorAggregatorConfig
from fme.ace.data_loading.batch_data import BatchData
from fme.ace.data_loading.getters import get_inference_data
from fme.ace.data_loading.inference import InferenceDataLoaderConfig
from fme.ace.inference.data_writer import DataWriterConfig, PairedDataWriter
from fme.ace.inference.data_writer.dataset_metadata import DatasetMetadata
from fme.ace.inference.data_writer.time_coarsen import TimeCoarsenConfig
from fme.ace.inference.default_metadata import get_default_variable_metadata
from fme.ace.inference.loop import DeriverABC, run_dataset_comparison
from fme.ace.stepper import (
    Stepper,
    StepperOverrideConfig,
    load_stepper,
    load_stepper_config,
)
from fme.ace.stepper.single_module import StepperConfig
from fme.core.cli import prepare_config, prepare_directory
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.dataset_info import IncompatibleDatasetInfo
from fme.core.derived_variables import get_derived_variable_metadata
from fme.core.dicts import to_flat_dict
from fme.core.generics.inference import get_record_to_wandb, run_inference
from fme.core.logging_utils import LoggingConfig
from fme.core.timing import GlobalTimer
from fme.core.typing_ import TensorDict, TensorMapping


def validate_time_coarsen_config(
    config: TimeCoarsenConfig, forward_steps_in_memory: int, n_forward_steps: int
):
    coarsen_factor = config.coarsen_factor
    if forward_steps_in_memory % coarsen_factor != 0:
        raise ValueError(
            "forward_steps_in_memory must be divisible by "
            f"time_coarsen.coarsen_factor. Got {forward_steps_in_memory} "
            f"and {coarsen_factor}."
        )
    if n_forward_steps % coarsen_factor != 0:
        raise ValueError(
            "n_forward_steps must be divisible by "
            f"time_coarsen.coarsen_factor. Got {n_forward_steps} "
            f"and {coarsen_factor}."
        )


def resolve_variable_metadata(
    dataset_metadata: Mapping[str, VariableMetadata],
    stepper_metadata: Mapping[str, VariableMetadata],
    stepper_all_names: Sequence[str],
) -> dict[str, VariableMetadata]:
    """
    Resolve variable metadata by merging from the following sources: derived variables,
    the dataset, the stepper, and finally a set of defaults. If there are conflicts on
    variable metadata values, preference is given first to values from the stepper,
    then from the dataset, and finally from default values.

    Note that if not saved with the stepper, the variable metadata is not guaranteed to
    be the same as that in the dataset used for training the stepper.

    Args:
        dataset_metadata: Metadata from the dataset.
        stepper_metadata: Metadata from the stepper.
        stepper_all_names: Variable names associated with the stepper.

    Returns:
        A mappping of variable names to metadata.
    """
    default_metadata = get_default_variable_metadata(version="era5_v1")
    names_from_default = (
        set(stepper_all_names) - (dataset_metadata.keys() | stepper_metadata.keys())
    ) & default_metadata.keys()
    if names_from_default:
        logging.warning(
            "Variable metadata for the following stepper variables were not found in "
            "the variable metadata of the forcing dataset or stepper: "
            f"{names_from_default}. Using default values for these variables instead. "
            "Users should ensure that the default values are consistent with the "
            "training dataset of the stepper."
        )
    resolved_metadata = (
        default_metadata | dict(dataset_metadata) | dict(stepper_metadata)
    )
    resolved_metadata = {
        name: resolved_metadata[name]
        for name in stepper_all_names
        if name in resolved_metadata
    }
    return get_derived_variable_metadata() | resolved_metadata


@dataclasses.dataclass
class InferenceEvaluatorConfig:
    """
    Configuration for running inference including comparison to reference data.

    Parameters:
        experiment_dir: Directory to save results to.
        n_forward_steps: Number of steps to run the model forward for.
        checkpoint_path: Path to stepper checkpoint to load.
        logging: configuration for logging.
        loader: Configuration for data to be used as initial conditions, forcing, and
            target in inference.
        prediction_loader: Configuration for prediction data to evaluate. If given,
            model evaluation will not run, and instead predictions will be evaluated.
            Model checkpoint will still be used to determine inputs and outputs.
        forward_steps_in_memory: Number of forward steps to complete in memory
            at a time, will load one more step for initial condition.
        data_writer: Configuration for data writers.
        aggregator: Configuration for inference evaluator aggregator.
        stepper_override: Configuration for overriding select stepper configuration
            options at inference time (optional).
        allow_incompatible_dataset: If True, allow the forcing dataset used
            for inference to be incompatible with the dataset used for stepper training.
            This should be used with caution, as it may allow the stepper to make
            scientifically invalid predictions, but it can allow running inference with
            incorrectly formatted or missing grid information.
    """

    experiment_dir: str
    n_forward_steps: int
    checkpoint_path: str
    logging: LoggingConfig
    loader: InferenceDataLoaderConfig
    prediction_loader: InferenceDataLoaderConfig | None = None
    forward_steps_in_memory: int = 1
    data_writer: DataWriterConfig = dataclasses.field(
        default_factory=lambda: DataWriterConfig()
    )
    aggregator: InferenceEvaluatorAggregatorConfig = dataclasses.field(
        default_factory=lambda: InferenceEvaluatorAggregatorConfig()
    )
    stepper_override: StepperOverrideConfig | None = None
    allow_incompatible_dataset: bool = False

    def __post_init__(self):
        if self.data_writer.time_coarsen is not None:
            validate_time_coarsen_config(
                self.data_writer.time_coarsen,
                self.forward_steps_in_memory,
                self.n_forward_steps,
            )

    def configure_logging(self, log_filename: str):
        self.logging.configure_logging(self.experiment_dir, log_filename)

    def configure_wandb(
        self, env_vars: dict | None = None, resumable: bool = False, **kwargs
    ):
        config = to_flat_dict(dataclasses.asdict(self))
        self.logging.configure_wandb(
            config=config, env_vars=env_vars, resumable=resumable, **kwargs
        )

    def load_stepper(self) -> Stepper:
        logging.info(f"Loading trained model checkpoint from {self.checkpoint_path}")
        return load_stepper(self.checkpoint_path, self.stepper_override)

    def load_stepper_config(self) -> StepperConfig:
        logging.info(f"Loading trained model checkpoint from {self.checkpoint_path}")
        return load_stepper_config(self.checkpoint_path, self.stepper_override)

    def get_data_writer(
        self,
        timestep: datetime.timedelta,
        variable_metadata: Mapping[str, VariableMetadata],
        coords: Mapping[str, np.ndarray],
    ) -> PairedDataWriter:
        return self.data_writer.build_paired(
            experiment_dir=self.experiment_dir,
            n_initial_conditions=self.loader.n_initial_conditions,
            n_timesteps=self.n_forward_steps,
            timestep=timestep,
            variable_metadata=variable_metadata,
            coords=coords,
            dataset_metadata=DatasetMetadata.from_env(),
        )


def main(yaml_config: str, override_dotlist: Sequence[str] | None = None):
    config_data = prepare_config(yaml_config, override=override_dotlist)
    config = dacite.from_dict(
        data_class=InferenceEvaluatorConfig,
        data=config_data,
        config=dacite.Config(strict=True),
    )
    prepare_directory(config.experiment_dir, config_data)
    with GlobalTimer():
        return run_evaluator_from_config(config)


class _Deriver(DeriverABC):
    """
    DeriverABC implementation for dataset comparison.
    """

    def __init__(
        self,
        n_ic_timesteps: int,
        derive_func: Callable[[TensorMapping, TensorMapping], TensorDict],
    ):
        self._n_ic_timesteps = n_ic_timesteps
        self._derive_func = derive_func

    @property
    def n_ic_timesteps(self) -> int:
        return self._n_ic_timesteps

    def get_forward_data(
        self, data: BatchData, compute_derived_variables: bool = False
    ) -> BatchData:
        if compute_derived_variables:
            timer = GlobalTimer.get_instance()
            with timer.context("compute_derived_variables"):
                data = data.compute_derived_variables(
                    derive_func=self._derive_func,
                    forcing_data=data,
                )
        return data.remove_initial_condition(self._n_ic_timesteps)


def run_evaluator_from_config(config: InferenceEvaluatorConfig):
    timer = GlobalTimer.get_instance()
    timer.start_outer("inference")
    timer.start("initialization")

    if not os.path.isdir(config.experiment_dir):
        os.makedirs(config.experiment_dir, exist_ok=True)
    config.configure_logging(log_filename="inference_out.log")
    env_vars = logging_utils.retrieve_env_vars()
    beaker_url = logging_utils.log_beaker_url()
    config.configure_wandb(env_vars=env_vars, notes=beaker_url)

    if fme.using_gpu():
        torch.backends.cudnn.benchmark = True

    logging_utils.log_versions()
    logging.info(f"Current device is {fme.get_device()}")

    stepper_config = config.load_stepper_config()
    logging.info("Initializing data loader")
    window_requirements = stepper_config.get_evaluation_window_data_requirements(
        n_forward_steps=config.forward_steps_in_memory
    )
    initial_condition_requirements = (
        stepper_config.get_prognostic_state_data_requirements()
    )
    data = get_inference_data(
        config=config.loader,
        total_forward_steps=config.n_forward_steps,
        window_requirements=window_requirements,
        initial_condition=initial_condition_requirements,
    )

    stepper = config.load_stepper()
    stepper.set_eval()

    if not config.allow_incompatible_dataset:
        try:
            stepper.training_dataset_info.assert_compatible_with(data.dataset_info)
        except IncompatibleDatasetInfo as err:
            raise IncompatibleDatasetInfo(
                "Inference dataset is not compatible with dataset used for stepper "
                "training. Set allow_incompatible_dataset to True to ignore this "
                f"error. The incompatiblity found was: {str(err)}"
            ) from err

    aggregator_config: InferenceEvaluatorAggregatorConfig = config.aggregator
    for batch in data.loader:
        initial_time = batch.time.isel(time=0)
        break
    variable_metadata = resolve_variable_metadata(
        dataset_metadata=data.variable_metadata,
        stepper_metadata=stepper.training_variable_metadata,
        stepper_all_names=stepper_config.all_names,
    )
    dataset_info = data.dataset_info.update_variable_metadata(variable_metadata)
    aggregator = aggregator_config.build(
        dataset_info=dataset_info,
        record_step_20=config.n_forward_steps >= 20,
        n_timesteps=config.n_forward_steps + stepper_config.n_ic_timesteps,
        initial_time=initial_time,
        channel_mean_names=stepper.loss_names,
        normalize=stepper.normalizer.normalize,
        output_dir=config.experiment_dir,
    )

    writer = config.get_data_writer(
        timestep=data.timestep,
        variable_metadata=variable_metadata,
        coords=data.coords,
    )

    timer.stop()
    logging.info("Starting inference")
    record_logs = get_record_to_wandb(label="inference")
    if config.prediction_loader is not None:
        prediction_data = get_inference_data(
            config.prediction_loader,
            total_forward_steps=config.n_forward_steps,
            window_requirements=window_requirements,
            initial_condition=initial_condition_requirements,
        )
        deriver = _Deriver(
            n_ic_timesteps=stepper_config.n_ic_timesteps,
            derive_func=stepper.derive_func,
        )
        run_dataset_comparison(
            aggregator=aggregator,
            prediction_data=prediction_data,
            target_data=data,
            deriver=deriver,
            writer=writer,
            record_logs=record_logs,
        )
    else:
        run_inference(
            predict=stepper.predict_paired,
            data=data,
            aggregator=aggregator,
            writer=writer,
            record_logs=record_logs,
        )

    timer.start("final_writer_flush")
    logging.info("Starting final flush of data writer")
    writer.flush()
    logging.info("Writing reduced metrics to disk in netcdf format.")
    aggregator.flush_diagnostics()
    timer.stop()

    timer.stop_outer("inference")
    total_steps = config.n_forward_steps * config.loader.n_initial_conditions
    inference_duration = timer.get_duration("inference")
    wandb_logging_duration = timer.get_duration("wandb_logging")
    total_steps_per_second = total_steps / (inference_duration - wandb_logging_duration)
    timer.log_durations()
    logging.info(
        "Total steps per second (ignoring wandb logging): "
        f"{total_steps_per_second:.2f} steps/second"
    )

    summary_logs = {
        "total_steps_per_second": total_steps_per_second,
        **timer.get_durations(),
        **aggregator.get_summary_logs(),
    }
    record_logs([summary_logs])



from collections.abc import Mapping

import torch
import xarray as xr

from fme.core.dataset.data_typing import VariableMetadata
from fme.core.distributed import Distributed
from fme.core.typing_ import TensorDict, TensorMapping
from fme.core.wandb import Image

from ..plotting import plot_paneled_data


class MapAggregator:
    """
    An aggregator that records the average over batches as function of lat and lon.
    """

    _captions = {
        "full-field": (
            "{name} one step mean full field; "
            "(top) generated and (bottom) target [{units}]"
        ),
        "error": (
            "{name} one step mean full field error (generated - target) [{units}]"
        ),
    }

    def __init__(
        self, dims: list[str], metadata: Mapping[str, VariableMetadata] | None = None
    ):
        """
        Args:
            dims: Dimensions of the data.
            metadata: Mapping of variable names their metadata that will
                used in generating logged image captions.
        """
        self._dims = dims
        if metadata is None:
            self._metadata: Mapping[str, VariableMetadata] = {}
        else:
            self._metadata = metadata
        self._n_batches = 0
        self._target_data: TensorDict = {}
        self._gen_data: TensorDict = {}

    @torch.no_grad()
    def record_batch(
        self,
        loss: float,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        target_data_norm: TensorMapping,
        gen_data_norm: TensorMapping,
    ):
        self._loss = loss
        for name in target_data:
            if name in self._target_data:
                self._target_data[name] += target_data[name].mean(dim=0)
            else:
                self._target_data[name] = target_data[name].mean(dim=0)
        for name in gen_data:
            if name in self._gen_data:
                self._gen_data[name] += gen_data[name].mean(dim=0)
            else:
                self._gen_data[name] = gen_data[name].mean(dim=0)
        self._n_batches += 1

    def _get_data(self) -> tuple[TensorMapping, TensorMapping]:
        dist = Distributed.get_instance()
        time_dim = 0
        # note that we are only using the first timestep
        # see https://github.com/ai2cm/full-model/issues/1005
        target_time = 1
        gen, target = {}, {}
        for name in sorted(list(self._gen_data.keys())):
            gen[name] = (
                (
                    dist.reduce_mean(
                        self._gen_data[name].select(dim=time_dim, index=target_time)
                    )
                    / self._n_batches
                )
                .cpu()
                .numpy()
            )
        for name in sorted(list(self._target_data.keys())):
            target[name] = (
                (
                    dist.reduce_mean(
                        self._target_data[name].select(dim=time_dim, index=target_time)
                    )
                    / self._n_batches
                )
                .cpu()
                .numpy()
            )
        return gen, target

    @torch.no_grad()
    def get_logs(self, label: str) -> dict[str, Image]:
        """
        Returns logs as can be reported to WandB.

        Args:
            label: Label to prepend to all log keys.
        """
        gen, target = self._get_data()
        image_logs = {}
        for name in gen.keys():
            image_logs[f"image-error/{name}"] = plot_paneled_data(
                [[(gen[name] - target[name])]],
                diverging=True,
                caption=self._get_caption("error", name),
            )
            image_logs[f"image-full-field/{name}"] = plot_paneled_data(
                [
                    [gen[name]],
                    [target[name]],
                ],
                diverging=False,
                caption=self._get_caption("full-field", name),
            )
        image_logs = {f"{label}/{key}": image_logs[key] for key in image_logs}
        return image_logs

    def _get_caption(self, key: str, name: str) -> str:
        if name in self._metadata:
            caption_name = self._metadata[name].long_name
            units = self._metadata[name].units
        else:
            caption_name, units = name, "unknown_units"
        caption = self._captions[key].format(name=caption_name, units=units)
        return caption

    def get_dataset(self) -> xr.Dataset:
        gen, target = self._get_data()
        ds = xr.Dataset()
        for name in gen:
            if name in self._metadata:
                long_name = self._metadata[name].long_name
                units = self._metadata[name].units
            else:
                long_name = name
                units = "unknown_units"
            metadata_attrs = {"long_name": long_name, "units": units}
            ds[f"gen_map-{name}"] = xr.DataArray(
                data=gen[name], dims=self._dims, attrs=metadata_attrs
            )
            ds[f"bias_map-{name}"] = xr.DataArray(
                data=gen[name] - target[name],
                dims=self._dims,
                attrs=metadata_attrs,
            )
        return ds



# flake8: noqa
# Copied from https://github.com/NVIDIA/modulus/commit/89a6091bd21edce7be4e0539cbd91507004faf08
# Copyright (c) 2023 - 2024 NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Sequence

import pandas as pd
import torch as th
import torch.nn as nn

from .healpix_decoder import UNetDecoderConfig
from .healpix_encoder import UNetEncoderConfig
from .healpix_layers import HEALPixFoldFaces, HEALPixUnfoldFaces


class HEALPixRecUNet(nn.Module):
    """Deep Learning Weather Prediction (DLWP) recurrent UNet model on the HEALPix mesh."""

    def __init__(
        self,
        encoder: UNetEncoderConfig,
        decoder: UNetDecoderConfig,
        input_channels: int,
        output_channels: int,
        prognostic_variables: int,
        n_constants: int,
        decoder_input_channels: int,
        input_time_size: int,
        output_time_size: int,
        delta_time: str = "6h",
        reset_cycle: str = "24h",
        presteps: int = 1,
        enable_nhwc: bool = False,
        enable_healpixpad: bool = False,
        couplings: list = [],
    ):
        """
        Initialize the HEALPixRecUNet model.

        Args:
            encoder: UNetEncoderConfig
                ModuleConfig of instantiable parameters for the U-net encoder.
            decoder: UNetDecoderConfig
                ModuleConfig of instantiable parameters for the U-net decoder.
            input_channels: int
                Number of input channels expected in the input array schema. Note this should be the
                number of input variables in the data, NOT including data reshaping for the encoder part.
            output_channels: int
                Number of output channels expected in the output array schema, or output variables.
            n_constants: int
                Number of optional constants expected in the input arrays. If this is zero, no constants
                should be provided as inputs to forward.
            decoder_input_channels: int
                Number of optional prescribed variables expected in the decoder input array
                for both inputs and outputs. If this is zero, no decoder inputs should be provided as inputs to forward.
            input_time_size: int
                Number of time steps in the input array.
            output_time_size: int
                Number of time steps in the output array.
            delta_time: str, optional
                Hours between two consecutive data points.
            reset_cycle: str, optional
                Hours after which the recurrent states are reset to zero and re-initialized. Set np.infty
                to never reset the hidden states.
            presteps: int, optional
                Number of model steps to initialize recurrent states.
            enable_nhwc: bool, optional
                Model with [N, H, W, C] instead of [N, C, H, W].
            enable_healpixpad: bool, optional
                Enable CUDA HEALPixPadding if installed.
            couplings: list, optional
                Sequence of dictionaries that describe coupling mechanisms. Currently unused in our production model;
                but we want to keep this in the module definition, in case we bring our SST module
                (which subclasses it) into the picture.
        """
        super().__init__()
        self.channel_dim = 2  # Now 2 with [B, F, T*C, H, W]. Was 1 in old data format with [B, T*C, F, H, W]

        self.input_channels = input_channels

        if n_constants == 0 and decoder_input_channels == 0:
            pass
            # raise NotImplementedError(
            #    "support for models with no constant fields and no decoder inputs (TOA insolation) is not available at this time."
            # )
        if couplings is not None:
            if len(couplings) > 0:
                if n_constants == 0:
                    raise NotImplementedError(
                        "support for coupled models with no constant fields is not available at this time."
                    )
                if decoder_input_channels == 0:
                    raise NotImplementedError(
                        "support for coupled models with no decoder inputs (TOA insolation) is not available at this time."
                    )
        else:
            couplings = []

        # add coupled fields to input channels for model initialization
        self.coupled_channels = self._compute_coupled_channels(couplings)
        self.couplings = couplings
        self.train_couplers = None
        self.output_channels = output_channels
        self.n_constants = n_constants
        self.prognostic_variables = prognostic_variables
        self.decoder_input_channels = decoder_input_channels
        self.input_time_size = input_time_size
        self.output_time_size = output_time_size
        self.delta_t = int(pd.Timedelta(delta_time).total_seconds() // 3600)
        self.reset_cycle = int(pd.Timedelta(reset_cycle).total_seconds() // 3600)
        self.presteps = presteps
        self.enable_nhwc = enable_nhwc
        self.enable_healpixpad = enable_healpixpad

        # Number of passes through the model, or a diagnostic model with only one output time
        self.is_diagnostic = self.output_time_size == 1 and self.input_time_size > 1
        if not self.is_diagnostic and (
            self.output_time_size % self.input_time_size != 0
        ):
            raise ValueError(
                f"'output_time_size' must be a multiple of 'input_time_size' (got "
                f"{self.output_time_size} and {self.input_time_size})"
            )

        # Build the model layers
        self.fold = HEALPixFoldFaces()
        self.unfold = HEALPixUnfoldFaces(num_faces=12)
        encoder.input_channels = self._compute_input_channels()
        encoder.enable_nhwc = self.enable_nhwc
        encoder.enable_healpixpad = self.enable_healpixpad
        self.encoder = encoder.build()

        self.encoder_depth = len(self.encoder.n_channels)
        decoder.output_channels = self._compute_output_channels()
        decoder.enable_nhwc = self.enable_nhwc
        decoder.enable_healpixpad = self.enable_healpixpad
        self.decoder = decoder.build()

    @property
    def integration_steps(self):
        """Number of integration steps"""
        return max(self.output_time_size // self.input_time_size, 1)

    def _compute_input_channels(self) -> int:
        """
        Calculate total number of input channels in the model.

        Returns:
            int: The total number of input channels.
        """
        return (
            self.input_time_size * (self.input_channels + self.decoder_input_channels)
            + self.n_constants
            + self.coupled_channels
        )

    def _compute_coupled_channels(self, couplings):
        """
        Get the number of coupled channels.

        Args:
            couplings: list
                Sequence of dictionaries that describe coupling mechanisms.

        Returns:
            int: The number of coupled channels.
        """
        c_channels = 0
        for c in couplings:
            c_channels += len(c["params"]["variables"]) * len(
                c["params"]["input_times"]
            )
        return c_channels

    def _compute_output_channels(self) -> int:
        """
        Compute the total number of output channels in the model.

        Returns:
            int: The total number of output channels.
        """
        return (
            1 if self.is_diagnostic else self.input_time_size
        ) * self.output_channels

    # def _reshape_inputs(self, inputs: Sequence, step: int = 0) -> th.Tensor:
    #     """
    #     Returns a single tensor to pass into the model encoder/decoder. Squashes the time/channel dimension and
    #     concatenates in constants and decoder inputs.

    #     Args:
    #         inputs: list of expected input tensors (inputs, decoder_inputs, constants)
    #         step: step number in the sequence of integration_steps

    #     Returns:
    #         reshaped Tensor in expected shape for model encoder
    #     """

    #     # if len(self.couplings) > 0:
    #     #     result = [
    #     #         inputs[0].flatten(
    #     #             start_dim=self.channel_dim, end_dim=self.channel_dim + 1
    #     #         ),
    #     #         inputs[1][
    #     #             :,
    #     #             :,
    #     #             slice(step * self.input_time_size, (step + 1) * self.input_time_size),
    #     #             ...,
    #     #         ].flatten(
    #     #             start_dim=self.channel_dim, end_dim=self.channel_dim + 1
    #     #         ),  # DI
    #     #         inputs[2].expand(
    #     #             *tuple([inputs[0].shape[0]] + len(inputs[2].shape) * [-1])
    #     #         ),  # constants
    #     #         inputs[3].permute(0, 2, 1, 3, 4),  # coupled inputs
    #     #     ]
    #     #     res = th.cat(result, dim=self.channel_dim)

    #     # else:
    #     #     if self.n_constants == 0:
    #     #         result = [  # This logic changes for no insolation layer for the time being
    #     #             inputs[0].flatten(
    #     #                 start_dim=self.channel_dim, end_dim=self.channel_dim + 1
    #     #             ),
    #     #             # inputs[0].flatten(
    #     #             #     start_dim=self.channel_dim, end_dim=self.channel_dim + 1
    #     #             # ),
    #     #             # inputs[1][
    #     #             #     :,
    #     #             #     :,
    #     #             #     slice(
    #     #             #         step * self.input_time_size, (step + 1) * self.input_time_size
    #     #             #     ),
    #     #             #     ...,
    #     #             # ].flatten(
    #     #             #     start_dim=self.channel_dim, end_dim=self.channel_dim + 1
    #     #             # ),  # DI
    #     #         ]
    #     #         res = th.cat(result, dim=self.channel_dim)

    #     #         # fold faces into batch dim
    #     #         res = self.fold(res)

    #     #         return res

    #     #     if self.decoder_input_channels == 0:
    #     #         result = [
    #     #             inputs[0].flatten(
    #     #                 start_dim=self.channel_dim, end_dim=self.channel_dim + 1
    #     #             ),
    #     #             inputs[1].expand(
    #     #                 *tuple([inputs[0].shape[0]] + len(inputs[1].shape) * [-1])
    #     #             ),  # inputs
    #     #         ]
    #     #         print(f"result 1 is {result[0].shape}")
    #     #         print(f"result 2 is {result[1].shape}")

    #     #         res = th.cat(result, dim=self.channel_dim)

    #     #         # fold faces into batch dim
    #     #         res = self.fold(res)

    #     #         return res

    #     #     result = [
    #     #         inputs[0].flatten(
    #     #             start_dim=self.channel_dim, end_dim=self.channel_dim + 1
    #     #         ),
    #     #         inputs[1][
    #     #             :,
    #     #             :,
    #     #             slice(step * self.input_time_size, (step + 1) * self.input_time_size),
    #     #             ...,
    #     #         ].flatten(
    #     #             start_dim=self.channel_dim, end_dim=self.channel_dim + 1
    #     #         ),  # DI
    #     #         inputs[2].expand(
    #     #             *tuple([inputs[0].shape[0]] + len(inputs[2].shape) * [-1])
    #     #         ),  # constants
    #     #     ]
    #     #     res = th.cat(result, dim=self.channel_dim)

    #     # fold faces into batch dim
    #     # res = self.fold(res)
    #     # res = self.fold(inputs)
    #     # return res

    # def _reshape_outputs(self, outputs: th.Tensor) -> th.Tensor:
    #     """Returns a multiple tensors to from the model decoder.
    #     Splits the time/channel dimensions.

    #     Args:
    #         inputs: list of expected input tensors (inputs, decoder_inputs, constants)
    #         step: step number in the sequence of integration_steps

    #     Returns:
    #         reshaped Tensor in expected shape for model outputs
    #     """
    #     # unfold:
    #     outputs = self.unfold(outputs)

    #     return outputs

    def _initialize_hidden(
        self, inputs: Sequence, outputs: Sequence, step: int
    ) -> None:
        """Initialize the hidden layers

        Args:
            inputs: Inputs to use to initialize the hideen layers
            outputs: Outputs to use to initialize the hideen layers
            step: Current step number of the initialization
        """
        self.reset()
        for prestep in range(self.presteps):
            if step < self.presteps:
                s = step + prestep
                if len(self.couplings) > 0:
                    input_tensor = self.fold(
                        inputs=[
                            inputs[0][
                                :,
                                :,
                                s * self.input_time_size : (s + 1)
                                * self.input_time_size,
                            ]
                        ]
                        + list(inputs[1:3])
                        + [inputs[3][prestep]],
                        step=step + prestep,
                    )
                else:
                    input_tensor = self.fold(
                        inputs=[
                            inputs[0][
                                :,
                                :,
                                s * self.input_time_size : (s + 1)
                                * self.input_time_size,
                            ]
                        ]
                        + list(inputs[1:]),
                        step=step + prestep,
                    )
            else:
                s = step - self.presteps + prestep
                if len(self.couplings) > 0:
                    input_tensor = self.fold(
                        inputs=[outputs[s - 1]]
                        + list(inputs[1:3])
                        + [inputs[3][step - (prestep - self.presteps)]],
                        step=s + 1,
                    )
                else:
                    input_tensor = self.fold(
                        inputs=[outputs[s - 1]] + list(inputs[1:]), step=s + 1
                    )
            self.decoder(self.encoder(input_tensor))

    def forward(self, inputs: th.Tensor, output_only_last=False) -> th.Tensor:
        """
        Forward pass of the HEALPixUnet

        Args:
            inputs: Inputs to the model, which is currently in the form [B, F, C, H, W].
                (We assume that constants have been preprocessed to persist across batch)
                (We also assume for now that the time is implied to be 1)

            Originally, this was expected to be a sequence of the form [prognostics|TISR|constants]:
                [B, F, T, C, H, W] the format for prognostics and TISR
                [F, C, H, W] the format for constants

            output_only_last: If only the last dimension of the outputs should
                be returned

        Returns:
            Predicted outputs
        """
        self.reset()  # will reset every step for now
        # We need to make sure that the new input is the correct size, since we no longer have the ability
        # to differentiate between the inputs, decoder inputs, and constants
        if self._compute_input_channels() != inputs.shape[2]:
            raise ValueError(
                f"Expected input should have channels {self._compute_input_channels()},"
                f" got {inputs.shape[2]}."
            )

        # The input logic gets really changed now that we just have a prognostic vars channel in the inputs.
        # Basically we assume that all the batch-wise expansion for the inputs has already been done.

        # (Re-)initialize recurrent hidden states
        # if (step * (self.delta_t * self.input_time_size)) % self.reset_cycle == 0:
        #     self._initialize_hidden(inputs=inputs, outputs=outputs, step=step)
        # Skipping this for now. We assume a single input time step, and resetting every step.
        # s = self.presteps
        input_tensor = self.fold(
            inputs
        )  # Padding happens in HEALPixPadding, which will
        # unfold this tensor to handle it.

        encodings = self.encoder(input_tensor)
        decodings = self.decoder(encodings)

        # Residual prediction
        n_prognostic_channels = self.prognostic_variables * self.input_time_size
        prognostic_outputs = (
            input_tensor[:, :n_prognostic_channels]
            + decodings[:, :n_prognostic_channels]
        )

        outputs_only = decodings[:, n_prognostic_channels:]

        reshaped = th.cat(
            [self.unfold(prognostic_outputs), self.unfold(outputs_only)],
            dim=self.channel_dim,
        )

        return reshaped

    def reset(self):
        """Resets the state of the network"""
        self.encoder.reset()
        self.decoder.reset()



import contextlib
import dataclasses
import itertools
from collections.abc import Callable, Iterable, Mapping
from typing import Any, Literal

import numpy as np
import torch
from torch import nn

from fme.core.device import get_device
from fme.core.generics.optimization import OptimizationABC
from fme.core.scheduler import SchedulerConfig
from fme.core.typing_ import TensorDict, TensorMapping


class Checkpoint:
    def __init__(self, kwargs: Mapping[str, Any]):
        self._kwargs = kwargs

    def __call__(self, module: nn.Module):
        def wrapped(*args):
            return torch.utils.checkpoint.checkpoint(
                module,
                *args,
                use_reentrant=False,
                **self._kwargs,
            )

        return wrapped


class NoCheckpoint:
    def __call__(self, module: nn.Module):
        return module


@dataclasses.dataclass
class CheckpointConfig:
    """
    Configuration for activation checkpointing.

    Trades increased computation in exchange for lowered memory consumption during
    training by recomputing activations in the backward pass.

    Parameters:
        after_n_forward_steps: Number of forward steps to generate before activation
            checkpointing is applied. Activation checkpointing is not used unless this
            number is less than the number of forward steps in the optimization.
        kwargs: Keyword arguments to pass to torch.utils.checkpoint.checkpoint.
            Note that use_reentrant=False is always explicitly passed
            as is recommended by the docs.
    """

    after_n_forward_steps: float = np.inf
    kwargs: Mapping[str, Any] = dataclasses.field(default_factory=dict)

    def build(self, step: int) -> Checkpoint | NoCheckpoint:
        """
        Builds a checkpoint function.

        Args:
            step: The current zero-indexed step number.

        Returns:
            A checkpoint function.
        """
        if step >= self.after_n_forward_steps:
            return Checkpoint(self.kwargs)
        else:
            return NoCheckpoint()


class Optimization(OptimizationABC):
    def __init__(
        self,
        parameters: Iterable[torch.nn.Parameter],
        optimizer_type: Literal["Adam", "FusedAdam"],
        lr: float,
        max_epochs: int,
        scheduler: SchedulerConfig,
        enable_automatic_mixed_precision: bool,
        kwargs: Mapping[str, Any],
        use_gradient_accumulation: bool = False,
        get_checkpoint: Callable[
            [int], Checkpoint | NoCheckpoint
        ] = lambda _: NoCheckpoint(),
    ):
        if optimizer_type == "FusedAdam":
            self.optimizer = torch.optim.AdamW(parameters, lr=lr, fused=True, **kwargs)
        elif optimizer_type == "Adam":
            self.optimizer = torch.optim.Adam(parameters, lr=lr, **kwargs)
        else:
            raise ValueError(f"Unknown optimizer type: {optimizer_type}")

        if enable_automatic_mixed_precision:
            self.gscaler: torch.amp.GradScaler | None = torch.amp.GradScaler("cuda")
        else:
            self.gscaler = None
        self.scheduler = scheduler.build(self.optimizer, max_epochs)
        self._accumulated_loss = torch.tensor(0.0, device=get_device())
        self._use_gradient_accumulation = use_gradient_accumulation
        self._get_checkpoint = get_checkpoint

    def checkpoint(self, module: nn.Module, step: int) -> nn.Module:
        return self._get_checkpoint(step)(module)

    @contextlib.contextmanager
    def autocast(self):
        enabled = self.gscaler is not None
        dtype = torch.bfloat16 if enabled else None
        with torch.amp.autocast("cuda", enabled=enabled, dtype=dtype):
            yield

    @property
    def learning_rate(self) -> float:
        return self.optimizer.param_groups[0]["lr"]

    def set_mode(self, modules: nn.ModuleList):
        """
        Sets the mode of the module to train.
        """
        for m in modules:
            m.train()

    def step_scheduler(self, valid_loss: float):
        """
        Step the scheduler.

        Args:
            valid_loss: The validation loss. Used in schedulers which change the
                learning rate based on whether the validation loss is decreasing.
        """
        if self.scheduler is not None:
            try:
                self.scheduler.step(metrics=valid_loss)
            except TypeError:
                self.scheduler.step()

    def detach_if_using_gradient_accumulation(self, state: TensorMapping) -> TensorDict:
        if self._use_gradient_accumulation:
            return {k: v.detach() for k, v in state.items()}
        return dict(state)

    def accumulate_loss(self, loss: torch.Tensor):
        self._validate_loss(loss)
        self._accumulated_loss += loss
        if self._use_gradient_accumulation:
            self._backward(loss)

    def get_accumulated_loss(self) -> torch.Tensor:
        return self._accumulated_loss

    def _backward(self, loss: torch.Tensor):
        if self.gscaler is not None:
            self.gscaler.scale(loss).backward()
        else:
            loss.backward()

    def _step_weights(self):
        if self.gscaler is not None:
            self.gscaler.step(self.optimizer)
        else:
            self.optimizer.step()

    def step_weights(self):
        if not self._use_gradient_accumulation:
            self._backward(self._accumulated_loss)
        self._step_weights()
        self.optimizer.zero_grad()
        if self.gscaler is not None:
            self.gscaler.update()
        self._accumulated_loss = torch.tensor(0.0, device=get_device())

    def get_state(self):
        """
        Returns state as a serializable data structure.
        """
        state = {
            "optimizer_state_dict": self.optimizer.state_dict(),
            "scheduler_state_dict": (
                self.scheduler.state_dict() if self.scheduler is not None else None
            ),
            "gscaler_state_dict": (
                self.gscaler.state_dict() if self.gscaler is not None else None
            ),
        }
        return state

    def load_state(self, state):
        """
        Loads state from a serializable data structure.
        """
        self.optimizer.load_state_dict(state["optimizer_state_dict"])
        if self.scheduler is not None:
            self.scheduler.load_state_dict(state["scheduler_state_dict"])
        if self.gscaler is not None:
            self.gscaler.load_state_dict(state["gscaler_state_dict"])

    def _validate_loss(self, loss: torch.Tensor):
        with torch.no_grad():
            if torch.isnan(loss):
                raise ValueError("Loss is NaN-valued during training.")


@dataclasses.dataclass
class OptimizationConfig:
    """
    Configuration for optimization.

    Parameters:
        optimizer_type: The type of optimizer to use.
        lr: The learning rate.
        kwargs: Additional keyword arguments to pass to the optimizer.
        enable_automatic_mixed_precision: Whether to use automatic mixed
            precision.
        scheduler: The type of scheduler to use. If none is given, no scheduler
            will be used.
        use_gradient_accumulation: Whether to use gradient accumulation. This must be
            supported by the stepper being optimized, which may accumulate gradients
            from separate losses to reduce memory consumption. The stepper may choose
            to accumulate gradients differently when this is enabled, such as by
            detaching the computational graph between steps. See the documentation of
            your stepper (e.g. Stepper) for more details.
    """

    optimizer_type: Literal["Adam", "FusedAdam"] = "Adam"
    lr: float = 0.001
    kwargs: Mapping[str, Any] = dataclasses.field(default_factory=dict)
    enable_automatic_mixed_precision: bool = False
    scheduler: SchedulerConfig = dataclasses.field(
        default_factory=lambda: SchedulerConfig()
    )
    use_gradient_accumulation: bool = False
    checkpoint: CheckpointConfig = dataclasses.field(
        default_factory=lambda: CheckpointConfig()
    )

    def build(self, modules: torch.nn.ModuleList, max_epochs: int) -> Optimization:
        parameters = itertools.chain(*[module.parameters() for module in modules])
        return Optimization(
            parameters=parameters,
            optimizer_type=self.optimizer_type,
            lr=self.lr,
            max_epochs=max_epochs,
            scheduler=self.scheduler,
            enable_automatic_mixed_precision=self.enable_automatic_mixed_precision,
            kwargs=self.kwargs,
            use_gradient_accumulation=self.use_gradient_accumulation,
            get_checkpoint=self.checkpoint.build,
        )

    def get_state(self) -> Mapping[str, Any]:
        return dataclasses.asdict(self)

    @classmethod
    def from_state(cls, state: Mapping[str, Any]) -> "OptimizationConfig":
        return cls(**state)


class NullOptimization(OptimizationABC):
    def __init__(self):
        self._accumulated_loss = torch.tensor(0.0, device=get_device())

    @contextlib.contextmanager
    def autocast(self):
        yield

    @property
    def learning_rate(self) -> float:
        return float("nan")

    def checkpoint(self, module: nn.Module, step: int) -> nn.Module:
        return module

    def step_scheduler(self, valid_loss: float):
        return

    def detach_if_using_gradient_accumulation(self, state: TensorMapping) -> TensorDict:
        return dict(state)

    def accumulate_loss(self, loss: torch.Tensor):
        self._accumulated_loss += loss

    def get_accumulated_loss(self) -> torch.Tensor:
        return self._accumulated_loss

    def step_weights(self):
        self._accumulated_loss = torch.tensor(0.0, device=get_device())
        return

    def get_state(self):
        return {}

    def load_state(self, state):
        return

    def set_mode(self, modules: nn.ModuleList):
        """
        Sets the mode of the module to eval.
        """
        for m in modules:
            m.eval()



import os
from collections.abc import Mapping
from typing import Protocol

import numpy as np
import xarray as xr

from fme.core.distributed import Distributed


class GetDataset(Protocol):
    def get_dataset(self) -> xr.Dataset: ...


def get_reduced_diagnostics(
    sub_aggregators: Mapping[str, GetDataset],
    coords: Mapping[str, np.ndarray],
) -> Mapping[str, xr.Dataset]:
    """
    Returns datasets from sub-aggregators. Coordinates are assigned to the datasets
    if they are present in the dataset.

    Args:
        sub_aggregators: dictionary of sub-aggregators.
        coords: dictionary of coordinates.

    Returns:
        Dictionary of datasets from aggregators.
    """
    datasets = {}
    for name, aggregator in sub_aggregators.items():
        ds = aggregator.get_dataset()
        ds_coords = {k: v for k, v in coords.items() if k in ds.dims}
        datasets[name] = ds.assign_coords(ds_coords)
    return datasets


def write_reduced_diagnostics(
    reduced_diagnostics: Mapping[str, xr.Dataset],
    output_dir: str,
    subdir: str | None = None,
):
    """Write the reduced metrics to disk. Each sub-aggregator will write a netCDF file
    if its `get_dataset` method returns a non-empty dataset.

    Args:
        reduced_diagnostics: Dictionary of reduced diagnostics datasets.
        output_dir: Output directory.
        subdir: Sub-directory within the output_dir.
    """
    if subdir is not None:
        output_dir = os.path.join(output_dir, subdir)
    dist = Distributed.get_instance()
    if dist.is_root():
        os.makedirs(output_dir, exist_ok=True)
        for name, ds in reduced_diagnostics.items():
            if len(ds) > 0:
                ds.to_netcdf(os.path.join(output_dir, f"{name}_diagnostics.nc"))



import dataclasses
import shutil
import tempfile
from typing import Dict, List

import click
import dacite
import fsspec
import xarray as xr
import yaml


def copy(source: str, destination: str):
    """Copy between any two 'filesystems'. Do not use for large files.

    Args:
        source: Path to source file/object.
        destination: Path to destination.
    """
    with fsspec.open(source) as f_source:
        with fsspec.open(destination, "wb") as f_destination:
            shutil.copyfileobj(f_source, f_destination)


@dataclasses.dataclass
class StatsConfig:
    output_directory: str
    exclude_runs: List[str] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class Config:
    runs: Dict[str, str]
    stats: StatsConfig


def add_history_attrs(ds, config_filename: str, stats_output_dir: str):
    ds.attrs["history"] = (
        "Created by full-model/fv3gfs_data_process/combine_stats.py from "
        f"configuration file {config_filename} using inputs at {stats_output_dir}."
    )


def open_datasets(roots: List[str], filename: str) -> List[xr.Dataset]:
    datasets = []
    for root in roots:
        with fsspec.open(root + filename) as file:
            ds = xr.open_dataset(file, decode_timedelta=False).load()
        datasets.append(ds)
    return datasets


@click.command()
@click.argument("config_yaml", type=str)
def main(config_yaml: str):
    """
    Combine statistics for the data processing pipeline.

    Arguments:
    config_yaml -- Path to the configuration file for the data processing pipeline.
    """
    with open(config_yaml, "r") as f:
        config_data = yaml.load(f, Loader=yaml.CLoader)
    config = dacite.from_dict(data_class=Config, data=config_data)
    xr.set_options(keep_attrs=True)

    stats_roots = [
        config.stats.output_directory + "/" + run + "/"
        for run in config.runs.keys()
        if run not in config.stats.exclude_runs
    ]
    with tempfile.TemporaryDirectory() as tmpdir:
        for filename in (
            "centering.nc",
            "scaling-residual.nc",
            "time-mean.nc",
        ):
            datasets = open_datasets(stats_roots, filename)
            if len(datasets) > 1:
                samples = xr.DataArray(
                    [float(ds.attrs["input_samples"]) for ds in datasets], dims=["run"]
                )
                combined = xr.concat(datasets, dim="run")
                if filename.startswith("scaling"):
                    # standard deviations are averaged as variances
                    average = (combined**2).weighted(samples).mean(dim="run") ** 0.5
                else:
                    average = combined.weighted(samples).mean(dim="run")
            else:
                average = datasets[0]

            add_history_attrs(average, config_yaml, config.stats.output_directory)
            average.to_netcdf(tmpdir + "/" + filename)
            copy(
                tmpdir + "/" + filename,
                config.stats.output_directory + "/combined/" + filename,
            )
        # for scaling-full-field.nc, we also need to account for the means.
        full_field_filename = "scaling-full-field.nc"
        centering_filename = "centering.nc"
        full_field_datasets = open_datasets(stats_roots, full_field_filename)
        if len(full_field_datasets) > 1:
            samples = xr.DataArray(
                [float(ds.attrs["input_samples"]) for ds in full_field_datasets],
                dims=["run"],
            )
            centering_datasets = open_datasets(stats_roots, centering_filename)
            average = get_combined_stats(
                full_field_datasets, centering_datasets, samples
            )
        else:
            average = full_field_datasets[0]

        add_history_attrs(average, config_yaml, config.stats.output_directory)
        average.to_netcdf(tmpdir + "/" + full_field_filename)
        copy(
            tmpdir + "/" + full_field_filename,
            config.stats.output_directory + "/combined/" + full_field_filename,
        )


def get_combined_stats(full_field_datasets, centering_datasets, samples):
    combined_scaling = xr.concat(full_field_datasets, dim="run")
    combined_centering = xr.concat(centering_datasets, dim="run")
    average_centering = combined_centering.weighted(samples).mean(dim="run")
    centering_variance = (combined_centering - average_centering) ** 2
    total_variance = centering_variance + combined_scaling**2
    average = total_variance.weighted(samples).mean(dim="run") ** 0.5
    return average


if __name__ == "__main__":
    main()



from .single_module import (
    ExistingStepperConfig,
    SingleModuleStepperConfig,
    Stepper,
    StepperOverrideConfig,
    TrainOutput,
    load_stepper,
    load_stepper_config,
    process_prediction_generator_list,
    stack_list_of_tensor_dicts,
)



import abc
from collections.abc import Callable
from typing import Any, TypeVar, final

import torch
import torch_harmonics

from fme.core import metrics
from fme.core.device import get_device
from fme.core.mask_provider import MaskProviderABC, NullMaskProvider
from fme.core.tensors import assert_dict_allclose
from fme.core.typing_ import TensorDict, TensorMapping


class GriddedOperations(abc.ABC):
    def __eq__(self, other) -> bool:
        if not isinstance(other, GriddedOperations):
            return False
        try:
            assert_dict_allclose(self.to_state(), other.to_state())
        except AssertionError:
            return False
        return True

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            + ", ".join(
                [f"{k}={v}" for k, v in self.get_initialization_kwargs().items()]
            )
            + ")"
        )

    @property
    @abc.abstractmethod
    def zonal_mean(self) -> Callable[[torch.Tensor], torch.Tensor] | None: ...

    @abc.abstractmethod
    def area_weighted_sum(
        self,
        data: torch.Tensor,
        keepdim: bool = False,
        name: str | None = None,
    ) -> torch.Tensor: ...

    @final
    def area_weighted_sum_dict(
        self, data: TensorMapping, keepdim: bool = False
    ) -> TensorDict:
        result = {}
        for name in data:
            result[name] = self.area_weighted_sum(
                data=data[name],
                keepdim=keepdim,
                name=name,
            )
        return result

    @abc.abstractmethod
    def area_weighted_mean(
        self,
        data: torch.Tensor,
        keepdim: bool = False,
        name: str | None = None,
    ) -> torch.Tensor: ...

    @final
    def area_weighted_mean_dict(
        self, data: TensorMapping, keepdim: bool = False
    ) -> TensorDict:
        result = {}
        for name in data:
            result[name] = self.area_weighted_mean(
                data=data[name],
                keepdim=keepdim,
                name=name,
            )
        return result

    def area_weighted_mean_bias(
        self,
        truth: torch.Tensor,
        predicted: torch.Tensor,
        name: str | None = None,
    ) -> torch.Tensor:
        return self.area_weighted_mean(predicted - truth, name=name)

    @final
    def area_weighted_mean_bias_dict(
        self, truth: TensorMapping, predicted: TensorMapping
    ) -> TensorDict:
        result = {}
        for name in truth:
            result[name] = self.area_weighted_mean_bias(
                truth=truth[name],
                predicted=predicted[name],
                name=name,
            )
        return result

    def area_weighted_rmse(
        self,
        truth: torch.Tensor,
        predicted: torch.Tensor,
        name: str | None = None,
    ) -> torch.Tensor:
        return torch.sqrt(self.area_weighted_mean((predicted - truth) ** 2, name=name))

    @final
    def area_weighted_rmse_dict(
        self, truth: TensorMapping, predicted: TensorMapping
    ) -> TensorDict:
        result = {}
        for name in truth:
            result[name] = self.area_weighted_rmse(
                truth=truth[name],
                predicted=predicted[name],
                name=name,
            )
        return result

    def area_weighted_std(
        self,
        data: torch.Tensor,
        keepdim: bool = False,
        name: str | None = None,
    ):
        return self.area_weighted_mean(
            (data - self.area_weighted_mean(data, keepdim=True, name=name)) ** 2,
            keepdim=keepdim,
            name=name,
        ).sqrt()

    @final
    def area_weighted_std_dict(
        self,
        data: TensorMapping,
        keepdim: bool = False,
    ) -> TensorDict:
        result = {}
        for name in data:
            result[name] = self.area_weighted_std(
                data=data[name],
                keepdim=keepdim,
                name=name,
            )
        return result

    @abc.abstractmethod
    def area_weighted_gradient_magnitude_percent_diff(
        self,
        truth: torch.Tensor,
        predicted: torch.Tensor,
        name: str | None = None,
    ): ...

    @final
    def area_weighted_gradient_magnitude_percent_diff_dict(
        self, truth: TensorMapping, predicted: TensorMapping
    ) -> TensorDict:
        result = {}
        for name in truth:
            result[name] = self.area_weighted_gradient_magnitude_percent_diff(
                truth=truth[name],
                predicted=predicted[name],
                name=name,
            )
        return result

    @abc.abstractmethod
    def regional_area_weighted_mean(
        self,
        data: torch.Tensor,
        regional_weights: torch.Tensor,
        keepdim: bool = False,
        name: str | None = None,
    ) -> torch.Tensor: ...

    @final
    def regional_area_weighted_mean_dict(
        self,
        data: TensorMapping,
        regional_weights: torch.Tensor,
        keepdim: bool = False,
    ) -> TensorDict:
        result = {}
        for name in data:
            result[name] = self.regional_area_weighted_mean(
                data=data[name],
                regional_weights=regional_weights,
                keepdim=keepdim,
                name=name,
            )
        return result

    @abc.abstractmethod
    def get_real_sht(
        self,
        grid: str = "legendre-gauss",
    ) -> torch_harmonics.RealSHT: ...

    def to_state(self) -> dict[str, Any]:
        return {
            "type": self.__class__.__name__,
            "state": self.get_initialization_kwargs(),
        }

    @abc.abstractmethod
    def get_initialization_kwargs(self) -> dict[str, Any]:
        """
        Get the keyword arguments needed to initialize the instance.
        """
        ...

    @classmethod
    def from_state(cls, state: dict[str, Any]) -> "GriddedOperations":
        """
        Given a dictionary with a "type" key and a "state" key, return
        the GriddedOperations it describes.

        The "type" key should be the name of a subclass of GriddedOperations,
        and the "state" key should be a dictionary specific to
        that subclass.

        Args:
            state: A dictionary with a "type" key and a "state" key.

        Returns:
            An instance of the subclass.
        """
        if cls is not GriddedOperations:
            raise RuntimeError(
                "This method should be called on GriddedOperations, "
                "not on its subclasses."
            )
        subclasses = get_all_subclasses(cls)
        for subclass in subclasses:
            if subclass.__name__ == state["type"]:
                return subclass(**state["state"])
        raise ValueError(
            f"Unknown subclass type: {state['type']}, "
            f"available: {[s.__name__ for s in subclasses]}"
        )


T = TypeVar("T")


def get_all_subclasses(cls: type[T]) -> list[type[T]]:
    """
    Gets all subclasses of a given class, including their subclasses etc.
    """
    all_subclasses = []
    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))
    return all_subclasses


def _mask_area_weights(
    area_weights: torch.Tensor,
    mask_provider: MaskProviderABC,
    name: str | None,
) -> torch.Tensor:
    if name is None:
        return area_weights
    mask = mask_provider.get_mask_tensor_for(name)
    if mask is None:
        return area_weights
    return area_weights * mask


class LatLonOperations(GriddedOperations):
    HORIZONTAL_DIMS = (-2, -1)

    def __init__(
        self,
        area_weights: torch.Tensor,
        mask_provider: MaskProviderABC = NullMaskProvider,
    ):
        # requires weights are longitudinally uniform
        if not torch.allclose(area_weights, area_weights[..., :1]):
            raise ValueError(
                "Area weights must be longitudinally uniform, "
                "as assumed for zonal mean."
            )
        self._device_area = area_weights.to(get_device())
        self._cpu_area = area_weights.to("cpu")
        self._device_mask_provider = mask_provider.to(get_device())
        self._cpu_mask_provider = mask_provider.to("cpu")

    @property
    def zonal_mean(self) -> Callable[[torch.Tensor], torch.Tensor]:
        return self._zonal_mean

    def _zonal_mean(self, data: torch.Tensor) -> torch.Tensor:
        return data.mean(dim=self.HORIZONTAL_DIMS[1])

    def _get_area_weights(
        self,
        data: torch.Tensor,
        name: str | None = None,
        regional_weights: torch.Tensor | None = None,
    ):
        if data.device == torch.device("cpu"):
            area_weights = self._cpu_area
            mask_provider = self._cpu_mask_provider
        else:
            area_weights = self._device_area
            mask_provider = self._device_mask_provider
        area_weights = _mask_area_weights(area_weights, mask_provider, name)
        if regional_weights is None:
            return area_weights
        if regional_weights.device.type != data.device.type:
            regional_weights = regional_weights.to(data.device)
        return regional_weights * area_weights

    def area_weighted_sum(
        self,
        data: torch.Tensor,
        keepdim: bool = False,
        name: str | None = None,
    ) -> torch.Tensor:
        area_weights = self._get_area_weights(data, name)
        return metrics.weighted_sum(
            data, area_weights, dim=self.HORIZONTAL_DIMS, keepdim=keepdim
        )

    def area_weighted_mean(
        self,
        data: torch.Tensor,
        keepdim: bool = False,
        name: str | None = None,
    ) -> torch.Tensor:
        area_weights = self._get_area_weights(data, name)
        return metrics.weighted_mean(
            data, area_weights, dim=self.HORIZONTAL_DIMS, keepdim=keepdim
        )

    def regional_area_weighted_mean(
        self,
        data: torch.Tensor,
        regional_weights: torch.Tensor,
        keepdim: bool = False,
        name: str | None = None,
    ) -> torch.Tensor:
        regional_area_weights = self._get_area_weights(data, name, regional_weights)
        return metrics.weighted_mean(
            data,
            regional_area_weights,
            dim=self.HORIZONTAL_DIMS,
            keepdim=keepdim,
        )

    def area_weighted_gradient_magnitude_percent_diff(
        self,
        truth: torch.Tensor,
        predicted: torch.Tensor,
        name: str | None = None,
    ):
        area_weights = self._get_area_weights(truth, name)
        return metrics.gradient_magnitude_percent_diff(
            truth,
            predicted,
            weights=area_weights,
            dim=self.HORIZONTAL_DIMS,
        )

    def get_real_sht(self, grid: str = "legendre-gauss") -> torch_harmonics.RealSHT:
        return torch_harmonics.RealSHT(
            self._cpu_area.shape[-2], self._cpu_area.shape[-1], grid=grid
        ).to(get_device())

    def get_initialization_kwargs(self) -> dict[str, Any]:
        return {"area_weights": self._cpu_area}


class HEALPixOperations(GriddedOperations):
    HORIZONTAL_DIMS = (-3, -2, -1)

    @property
    def zonal_mean(self) -> None:
        # not implemented, though we definitely could
        # as HEALPix rings are constant-latitude
        return None

    def area_weighted_sum(
        self,
        data: torch.Tensor,
        keepdim: bool = False,
        name: str | None = None,
    ) -> torch.Tensor:
        # For HEALPix, area weights are uniform, so sum is sufficient
        return data.sum(dim=self.HORIZONTAL_DIMS, keepdim=keepdim)

    def area_weighted_mean(
        self,
        data: torch.Tensor,
        keepdim: bool = False,
        name: str | None = None,
    ) -> torch.Tensor:
        # For HEALPix, area weights are uniform, so mean is sufficient
        return data.mean(dim=self.HORIZONTAL_DIMS, keepdim=keepdim)

    def area_weighted_gradient_magnitude_percent_diff(
        self,
        truth: torch.Tensor,
        predicted: torch.Tensor,
        name: str | None = None,
    ) -> torch.Tensor:
        return metrics.gradient_magnitude_percent_diff(
            truth, predicted, weights=None, dim=self.HORIZONTAL_DIMS
        )

    def regional_area_weighted_mean(
        self,
        data: torch.Tensor,
        weights: torch.Tensor,
        keepdim: bool = False,
        name: str | None = None,
    ) -> torch.Tensor:
        raise NotImplementedError(
            "Regional area weighted mean is not implemented for HEALPix."
        )

    def get_real_sht(self, grid: str = "legendre-gauss") -> torch_harmonics.RealSHT:
        raise NotImplementedError("SHT is not implemented for HEALPix.")

    def get_initialization_kwargs(self) -> dict[str, Any]:
        return {}



import sys

from fme.ace.aggregator.one_step import OneStepAggregatorConfig
from fme.ace.data_loading.augmentation import AugmentationConfig
from fme.ace.data_loading.inference import (
    ExplicitIndices,
    InferenceInitialConditionIndices,
    TimestampList,
)
from fme.ace.data_loading.perturbation import (
    ConstantConfig,
    GreensFunctionConfig,
    PerturbationSelector,
    SSTPerturbation,
)
from fme.ace.inference.data_writer.time_coarsen import TimeCoarsenConfig
from fme.ace.inference.evaluator import (
    DataWriterConfig,
    InferenceDataLoaderConfig,
    InferenceEvaluatorAggregatorConfig,
    InferenceEvaluatorConfig,
    run_evaluator_from_config,
)
from fme.ace.inference.inference import (
    ForcingDataLoaderConfig,
    InferenceAggregatorConfig,
    InferenceConfig,
    InitialConditionConfig,
    run_inference_from_config,
)
from fme.ace.models.healpix.healpix_activations import (
    CappedGELUConfig,
    DownsamplingBlockConfig,
)
from fme.ace.models.healpix.healpix_blocks import ConvBlockConfig, RecurrentBlockConfig
from fme.ace.registry.hpx import (
    HEALPixRecUNetBuilder,
    UNetDecoderConfig,
    UNetEncoderConfig,
)
from fme.ace.registry.sfno import SFNO_V0_1_0, SphericalFourierNeuralOperatorBuilder
from fme.ace.stepper import StepperOverrideConfig
from fme.ace.stepper.parameter_init import (
    FrozenParameterConfig,
    ParameterInitializationConfig,
)
from fme.ace.stepper.single_module import StepperConfig
from fme.core.corrector.atmosphere import AtmosphereCorrectorConfig
from fme.core.corrector.ocean import OceanCorrectorConfig
from fme.core.dataset.config import (
    ConcatDatasetConfig,
    FillNaNsConfig,
    MergeDatasetConfig,
    MergeNoConcatDatasetConfig,
    OverwriteConfig,
    RepeatedInterval,
    TimeSlice,
    XarrayDataConfig,
)
from fme.core.gridded_ops import GriddedOperations
from fme.core.loss import WeightedMappingLossConfig
from fme.core.multi_call import MultiCallConfig
from fme.core.normalizer import NormalizationConfig
from fme.core.ocean import OceanConfig, SlabOceanConfig
from fme.core.optimization import SchedulerConfig
from fme.core.registry.corrector import CorrectorSelector
from fme.core.registry.module import ModuleSelector
from fme.core.typing_ import Slice

from .train.train import run_train
from .train.train_config import (
    CopyWeightsConfig,
    DataLoaderConfig,
    EMAConfig,
    ExistingStepperConfig,
    InlineInferenceConfig,
    LoggingConfig,
    OptimizationConfig,
    SingleModuleStepperConfig,
    TrainConfig,
)

# Get all the names defined in the current module
module = sys.modules[__name__]
__all__ = [name for name in dir(module) if not name.startswith("_")]
del sys, module
